# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Testing :mod:`astropy.cosmology.w0wzcdm`."""

import numpy as np
import pytest

import astropy.units as u
from astropy.cosmology import Flatw0wzCDM, w0wzCDM
from astropy.cosmology._src.parameter import Parameter
from astropy.cosmology._src.tests.test_core import ParameterTestMixin, make_valid_zs
from astropy.utils.compat.optional_deps import HAS_SCIPY

from .conftest import filter_keys_from_items
from .test_base import FlatFLRWMixinTest, FLRWTest
from .test_w0cdm import Parameterw0TestMixin

##############################################################################
# PARAMETERS

COMOVING_DISTANCE_EXAMPLE_KWARGS = {"w0": -0.9, "wz": 0.1, "Tcmb0": 0.0}

valid_zs = make_valid_zs(max_z=400)[-1]


##############################################################################
# TESTS
##############################################################################


class ParameterwzTestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` wz on a Cosmology.

    wz is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_wz(self, cosmo_cls, cosmo):
        """Test Parameter ``wz``."""
        # on the class
        wz = cosmo_cls.parameters["wz"]
        assert isinstance(wz, Parameter)
        assert "Derivative of the dark energy" in wz.__doc__
        assert wz.unit is None
        assert wz.default == 0.0

        # on the instance
        assert cosmo.wz is cosmo.__dict__["wz"]
        assert cosmo.wz == self.cls_kwargs["wz"]

    def test_init_wz(self, cosmo_cls, ba):
        """Test initialization for values of ``wz``."""
        # test that it works with units
        ba.arguments["wz"] = ba.arguments["wz"] << u.one  # ensure units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.wz == ba.arguments["wz"]

        # also without units
        ba.arguments["wz"] = ba.arguments["wz"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.wz == ba.arguments["wz"]

        # must be dimensionless
        ba.arguments["wz"] = 10 * u.km
        with pytest.raises(TypeError):
            cosmo_cls(*ba.args, **ba.kwargs)


class Testw0wzCDM(FLRWTest, Parameterw0TestMixin, ParameterwzTestMixin):
    """Test :class:`astropy.cosmology.w0wzCDM`."""

    def setup_class(self):
        """Setup for testing."""
        super().setup_class(self)
        self.cls = w0wzCDM
        self.cls_kwargs.update(w0=-1, wz=0.5)

    # ===============================================================
    # Method & Attribute Tests

    def test_clone_change_param(self, cosmo):
        """Test method ``.clone()`` changing a(many) Parameter(s)."""
        super().test_clone_change_param(cosmo)

        # `w` params
        c = cosmo.clone(w0=0.1, wz=0.2)
        assert c.w0 == 0.1
        assert c.wz == 0.2
        for n, v in filter_keys_from_items(c.parameters, ("w0", "wz")):
            assert u.allclose(v, getattr(cosmo, n), atol=1e-4 * getattr(v, "unit", 1))

    # @pytest.mark.parametrize("z", valid_zs)  # TODO! recompute comparisons below
    def test_w(self, cosmo):
        """Test :meth:`astropy.cosmology.w0wzCDM.w`."""
        # super().test_w(cosmo, z)

        assert u.allclose(cosmo.w(1.0), -0.5)
        assert u.allclose(
            cosmo.w([0.0, 0.5, 1.0, 1.5, 2.3]), [-1.0, -0.75, -0.5, -0.25, 0.15]
        )

    def test_repr(self, cosmo_cls, cosmo):
        """Test method ``.__repr__()``."""
        assert repr(cosmo) == (
            "w0wzCDM(name='ABCMeta', H0=<Quantity 70. km / (Mpc s)>, Om0=0.27, "
            "Ode0=0.73, Tcmb0=<Quantity 3. K>, Neff=3.04, "
            "m_nu=<Quantity [0., 0., 0.] eV>, Ob0=0.03, w0=-1.0, wz=0.5)"
        )

    # ---------------------------------------------------------------

    @pytest.mark.parametrize("z", valid_zs)
    def test_Otot(self, cosmo, z):
        """Test :meth:`astropy.cosmology.w0wzCDM.Otot`.

        This is tested in the base class, but we need to override it here because
        this class is quite unstable.
        """
        super().test_Otot(cosmo, z)

    def test_Otot_overflow(self, cosmo):
        """Test :meth:`astropy.cosmology.w0wzCDM.Otot` for overflow."""
        with (
            np.errstate(invalid="ignore", over="warn"),
            pytest.warns(RuntimeWarning, match="overflow encountered in exp"),
        ):
            cosmo.Otot(1e3)

    # ===============================================================
    # I/O Tests

    @pytest.mark.filterwarnings("ignore:overflow encountered")
    def test_toformat_model(self, cosmo, to_format, method_name):
        """Test cosmology -> astropy.model."""
        super().test_toformat_model(cosmo, to_format, method_name)

    # ===============================================================
    # Usage Tests

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy is not installed")
    @pytest.mark.parametrize(
        ("args", "kwargs", "expected"),
        [
            (  # no relativistic species
                (75.0, 0.3, 0.6),
                {},
                [2934.20187523, 4559.94636182, 5590.71080419, 6312.66783729] * u.Mpc,
            ),
            (  # massless neutrinos
                (75.0, 0.25, 0.5),
                {"Tcmb0": 3.0, "Neff": 3, "m_nu": 0 * u.eV},
                [2904.47062713, 4528.59073707, 5575.95892989, 6318.98689566] * u.Mpc,
            ),
            (  # massive neutrinos
                (75.0, 0.25, 0.5),
                {"Tcmb0": 3.0, "Neff": 4, "m_nu": 5 * u.eV},
                [2613.84726408, 3849.66574595, 4585.51172509, 5085.16795412] * u.Mpc,
            ),
        ],
    )
    def test_comoving_distance_example(self, cosmo_cls, args, kwargs, expected):
        """Test :meth:`astropy.cosmology.LambdaCDM.comoving_distance`.

        These do not come from external codes -- they are just internal checks to make
        sure nothing changes if we muck with the distance calculators.
        """
        super().test_comoving_distance_example(
            cosmo_cls, args, {**COMOVING_DISTANCE_EXAMPLE_KWARGS, **kwargs}, expected
        )

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy is not installed")
    def test_comoving_distance_mathematica(self, cosmo_cls):
        """Test with Mathematica example.

        This test should be updated as the code changes.

        ::
            In[1]:= {Om0, w0, wz, H0, c}={0.3,-0.9, 0.2, 70, 299792.458};
            c/H0 NIntegrate[1/Sqrt[Om0*(1+z)^3+(1-Om0)(1+z)^(3(1+w0-wz)) Exp[3 *wz*z]],{z, 0, 0.5}]
            Out[1]= 1849.75
        """
        assert u.allclose(
            cosmo_cls(H0=70, Om0=0.3, w0=-0.9, wz=0.2, Ode0=0.7).comoving_distance(0.5),
            1849.75 * u.Mpc,
            rtol=1e-4,
        )


class TestFlatw0wzCDM(FlatFLRWMixinTest, Testw0wzCDM):
    """Test :class:`astropy.cosmology.Flatw0wzCDM`."""

    def setup_class(self):
        """Setup for testing."""
        super().setup_class(self)
        self.cls = Flatw0wzCDM

    def test_repr(self, cosmo_cls, cosmo):
        """Test method ``.__repr__()``."""
        super().test_repr(cosmo_cls, cosmo)

        assert repr(cosmo) == (
            "Flatw0wzCDM(name='ABCMeta', H0=<Quantity 70. km / (Mpc s)>, Om0=0.27, "
            "Tcmb0=<Quantity 3. K>, Neff=3.04, m_nu=<Quantity [0., 0., 0.] eV>, "
            "Ob0=0.03, w0=-1.0, wz=0.5)"
        )

    # ---------------------------------------------------------------

    @pytest.mark.parametrize("z", valid_zs)
    def test_Otot(self, cosmo, z):
        """Test :meth:`astropy.cosmology.Flatw0wzCDM.Otot`.

        This is tested in the base class, but we need to override it here because
        this class is quite unstable.
        """
        super().test_Otot(cosmo, z)

    def test_Otot_overflow(self, cosmo):
        """Test :meth:`astropy.cosmology.Flatw0wzCDM.Otot` for NOT overflowing."""
        cosmo.Otot(1e5)

    # ---------------------------------------------------------------

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy is not installed")
    @pytest.mark.parametrize(
        ("args", "kwargs", "expected"),
        [
            (  # no relativistic species
                (75.0, 0.3),
                {},
                [3004.55645039, 4694.15295565, 5760.90038238, 6504.07869144] * u.Mpc,
            ),
            (  # massless neutrinos
                (75.0, 0.25),
                {"Tcmb0": 3.0, "Neff": 3, "m_nu": 0 * u.eV},
                [3086.14574034, 4885.09170925, 6035.4563298, 6840.89215656] * u.Mpc,
            ),
            (  # massive neutrinos
                (75.0, 0.25),
                {"Tcmb0": 3.0, "Neff": 4, "m_nu": 5 * u.eV},
                [2510.44035219, 3683.87910326, 4389.97760294, 4873.33577288] * u.Mpc,
            ),
        ],
    )
    def test_comoving_distance_example(self, cosmo_cls, args, kwargs, expected):
        """Test :meth:`astropy.cosmology.LambdaCDM.comoving_distance`.

        These do not come from external codes -- they are just internal checks to make
        sure nothing changes if we muck with the distance calculators.
        """
        super().test_comoving_distance_example(
            cosmo_cls, args, {**COMOVING_DISTANCE_EXAMPLE_KWARGS, **kwargs}, expected
        )

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy is not installed")
    def test_comoving_distance_mathematica(self, cosmo_cls):
        """Test with Mathematica example.

        This test should be updated as the code changes.

        ::
            In[1]:= {Om0, w0, wz, H0, c}={0.3,-0.9, 0.2, 70, 299792.458};
            c/H0 NIntegrate[1/Sqrt[Om0*(1+z)^3+(1-Om0)(1+z)^(3(1+w0-wz)) Exp[3 *wz*z]],{z, 0, 0.5}]
            Out[1]= 1849.75
        """
        assert u.allclose(
            cosmo_cls(H0=70, Om0=0.3, w0=-0.9, wz=0.2).comoving_distance(0.5),
            1849.75 * u.Mpc,
            rtol=1e-4,
        )


##############################################################################
# Miscellaneous
# TODO: these should be better integrated into the new test framework


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
def test_de_densityscale():
    cosmo = w0wzCDM(H0=70, Om0=0.3, Ode0=0.50, w0=-1, wz=0.5)

    z = np.array([0.1, 0.2, 0.5, 1.5, 2.5])
    assert u.allclose(
        cosmo.de_density_scale(z),
        [1.00705953, 1.02687239, 1.15234885, 2.40022841, 6.49384982],
        rtol=1e-4,
    )

    assert u.allclose(cosmo.de_density_scale(3), cosmo.de_density_scale(3.0), rtol=1e-7)
    assert u.allclose(
        cosmo.de_density_scale([1, 2, 3]),
        cosmo.de_density_scale([1.0, 2.0, 3.0]),
        rtol=1e-7,
    )

    # Flat tests
    cosmo = w0wzCDM(H0=70, Om0=0.3, Ode0=0.7, w0=-1, wz=0.5)
    flatcosmo = Flatw0wzCDM(H0=70, Om0=0.3, w0=-1, wz=0.5)

    assert u.allclose(
        cosmo.de_density_scale(z), flatcosmo.de_density_scale(z), rtol=1e-4
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from astropy.io.ascii import read

# NOTE: Python can be built without bz2 or lzma
from astropy.utils.compat.optional_deps import HAS_BZ2, HAS_LZMA, HAS_UNCOMPRESSPY
from astropy.utils.data import get_pkg_data_filename


@pytest.mark.parametrize(
    "filename", ["data/daophot.dat.gz", "data/latex1.tex.gz", "data/short.rdb.gz"]
)
def test_gzip(filename):
    t_comp = read(get_pkg_data_filename(filename))
    t_uncomp = read(get_pkg_data_filename(filename.replace(".gz", "")))
    assert t_comp.dtype.names == t_uncomp.dtype.names
    assert np.all(t_comp.as_array() == t_uncomp.as_array())


@pytest.mark.xfail(not HAS_BZ2, reason="requires bz2")
@pytest.mark.parametrize("filename", ["data/short.rdb.bz2", "data/ipac.dat.bz2"])
def test_bzip2(filename):
    t_comp = read(get_pkg_data_filename(filename))
    t_uncomp = read(get_pkg_data_filename(filename.replace(".bz2", "")))
    assert t_comp.dtype.names == t_uncomp.dtype.names
    assert np.all(t_comp.as_array() == t_uncomp.as_array())


@pytest.mark.xfail(not HAS_LZMA, reason="requires lzma")
@pytest.mark.parametrize("filename", ["data/short.rdb.xz", "data/ipac.dat.xz"])
def test_xz(filename):
    t_comp = read(get_pkg_data_filename(filename))
    t_uncomp = read(get_pkg_data_filename(filename.replace(".xz", "")))
    assert t_comp.dtype.names == t_uncomp.dtype.names
    assert np.all(t_comp.as_array() == t_uncomp.as_array())


@pytest.mark.xfail(not HAS_UNCOMPRESSPY, reason="requires uncompresspy")
@pytest.mark.parametrize("filename", ["data/short.rdb.Z", "data/ipac.dat.Z"])
def test_lzw(filename):
    t_comp = read(get_pkg_data_filename(filename))
    t_uncomp = read(get_pkg_data_filename(filename.removesuffix(".Z")))
    assert t_comp.dtype.names == t_uncomp.dtype.names
    assert_array_equal(t_comp.as_array(), t_uncomp.as_array())



import copy

import numpy as np
import pytest

import astropy.units as u
from astropy.io.misc.parquet import read_parquet_votable, write_parquet_votable
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning
from astropy.utils.misc import _NOT_OVERWRITING_MSG_MATCH

# Skip all tests in this file if we cannot import pyarrow
pyarrow = pytest.importorskip("pyarrow")


number_of_objects = 10
ids = [f"COSMOS_{ii}" for ii in range(number_of_objects)]
redshift = np.random.uniform(low=0, high=3, size=number_of_objects)
mass = np.random.uniform(low=1e8, high=1e10, size=number_of_objects)
sfr = np.random.uniform(low=1, high=100, size=number_of_objects)

input_table = Table(
    [ids, redshift, mass, sfr],
    names=["id", "z", "mass", "sfr"],
)

column_metadata = {
    "id": {
        "unit": "",
        "ucd": "meta.id",
        "utype": "none",
        "description": "The ID of the galaxy.",
    },
    "z": {
        "unit": "",
        "ucd": "src.redshift",
        "utype": "none",
        "description": "The redshift of the galaxy.",
    },
    "mass": {
        "unit": "solMass",
        "ucd": "phys.mass",
        "utype": "none",
        "description": "The stellar mass of the galaxy.",
    },
    "sfr": {
        "unit": "solMass / yr",
        "ucd": "phys.SFR",
        "utype": "none",
        "description": "The star formation rate of the galaxy.",
    },
}


def test_parquet_votable(tmp_path):
    """Test writing and reading a votable into a parquet file"""
    filename = tmp_path / "test_votable.parq"

    write_parquet_votable(input_table, filename, metadata=column_metadata)

    with pytest.warns(AstropyUserWarning, match="No table::len"):
        loaded_table = read_parquet_votable(filename)

    # Compare data content, but not metadata (input had none)
    assert np.all(input_table == loaded_table)


@pytest.mark.parametrize("overwrite_metadata", [True, False])
def test_parquet_votable_input_column_unit(overwrite_metadata, tmp_path):
    """Test round trip of a table that has column units"""
    filename = tmp_path / "test_votable.parq"

    modified_input = copy.deepcopy(input_table)
    modified_input["mass"].unit = u.kg

    write_parquet_votable(
        modified_input,
        filename,
        metadata=column_metadata,
        overwrite_metadata=overwrite_metadata,
    )

    with pytest.warns(AstropyUserWarning, match="No table::len"):
        loaded_table = read_parquet_votable(filename)

    # Compare data content
    assert np.all(modified_input == loaded_table)

    assert (
        modified_input["mass"].unit == loaded_table["mass"].unit
    ) is not overwrite_metadata
    assert loaded_table["sfr"].unit == u.solMass / u.yr


@pytest.mark.xfail(reason="TODO fix: existing column metadata is ignored")
def test_parquet_votable_input_column_metadata(tmp_path):
    """Test preservation of column metadata"""
    filename = tmp_path / "test_votable.parq"

    modified_input = copy.deepcopy(input_table)
    modified_input["mass"].meta["foo"] = "bar"
    write_parquet_votable(modified_input, filename, metadata=column_metadata)

    with pytest.warns(AstropyUserWarning, match="No table::len"):
        loaded_table = read_parquet_votable(filename)

    assert "foo" in loaded_table["mass"].meta
    assert loaded_table["mass"].meta["foo"] == "bar"


@pytest.mark.xfail(reason="TODO fix: existing metadata is ignored")
def test_parquet_votable_input_metadata(tmp_path):
    """Test preservation of table metadata"""
    filename = tmp_path / "test_votable.parq"

    modified_input = copy.deepcopy(input_table)
    modified_input.meta["table_foo"] = "table_bar"
    write_parquet_votable(modified_input, filename, metadata=column_metadata)

    with pytest.warns(AstropyUserWarning, match="No table::len"):
        loaded_table = read_parquet_votable(filename)

    assert "table_foo" in loaded_table.meta
    assert loaded_table.meta["table_foo"] == "table_bar"


def test_compare_parquet_votable(tmp_path):
    """'parquet.votable' preserves column units and metadata unlike 'parquet'"""

    filename = tmp_path / "test_votable.parq"
    write_parquet_votable(input_table, filename, metadata=column_metadata)

    with pytest.warns(AstropyUserWarning, match="No table::len"):
        parquet_table = Table.read(filename, format="parquet")
        parquet_votable = Table.read(filename, format="parquet.votable")

    assert len(parquet_table["sfr"].meta) == 0
    assert parquet_table["sfr"].unit is None

    assert parquet_votable["sfr"].meta["ucd"] == "phys.SFR"
    assert parquet_votable["sfr"].unit == u.solMass / u.yr


def test_write_from_votable_existing_metadata(tmp_path):
    """Read a votable into a Table and write it out as 'parquet.votable' preserving metadata"""
    output_filename = tmp_path / "test_votable.parq"
    input_data = get_pkg_data_filename("data/gaia_source_dr3_select_1_result.vot")

    input_t = Table.read(input_data)

    with pytest.raises(NotImplementedError):
        # Write out the VOTable into a parquet, preserving the metadata
        # without providing an extra dictionary
        input_t.write(output_filename, format="parquet.votable")

        loaded_table = Table.read(output_filename, format="parquet.votable")

        assert np.all(input_t == loaded_table)
        for attr in ["meta", "unit", "dtype", "description"]:
            assert getattr(input_t["ra"], attr) == getattr(loaded_table["ra"], attr)


# Not sure this use case should be supported at all
@pytest.mark.xfail(reason="TODO fix: metadata could be overwritten")
def test_write_from_votable_overriding_metadata(tmp_path):
    """Read a votable into a Table and write it out as 'parquet.votable' preserving metadata"""
    output_filename = tmp_path / "test_votable.parq"
    input_data = get_pkg_data_filename("data/gaia_source_dr3_select_1_result.vot")

    input_t = Table.read(input_data)

    new_metadata = {
        "SOURCE_ID": {
            "description": "Updated description",
        },
    }
    # Write out the VOTable into a parquet, preserving the metadata
    # without providing an extra dictionary
    input_t.write(
        output_filename,
        format="parquet.votable",
        metadata=new_metadata,
        overwrite_metadata=True,
    )

    loaded_table = Table.read(output_filename, format="parquet.votable")

    assert loaded_table["SOURCE_ID"].description == "Updated description"


def test_read_write_existing(tmp_path):
    """Test writing an existing file without overwriting."""
    filename = tmp_path / "test_votable.parquet"
    with open(filename, "w"):
        # create empty file
        pass

    with pytest.raises(OSError, match=_NOT_OVERWRITING_MSG_MATCH):
        write_parquet_votable(input_table, filename, metadata=column_metadata)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest

from astropy.io import fits
from astropy.utils.compat.optional_deps import HAS_MATPLOTLIB

if HAS_MATPLOTLIB:
    import matplotlib.image as mpimg

    from astropy.visualization.scripts.fits2bitmap import fits2bitmap, main


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason="requires matplotlib")
class TestFits2Bitmap:
    def setup_class(self):
        self.filename = "test.fits"
        self.array = np.arange(16384).reshape((128, 128))

    def test_function(self, tmp_path):
        filename = tmp_path / self.filename
        fits.writeto(filename, self.array)
        fits2bitmap(filename)

    def test_script(self, tmp_path):
        filename = str(tmp_path / self.filename)
        fits.writeto(filename, self.array)
        main([filename, "-e", "0"])

    def test_exten_num(self, tmp_path):
        filename = str(tmp_path / self.filename)
        hdu1 = fits.PrimaryHDU()
        hdu2 = fits.ImageHDU(self.array)
        hdulist = fits.HDUList([hdu1, hdu2])
        hdulist.writeto(filename)
        main([filename, "-e", "1"])

    def test_exten_name(self, tmp_path):
        filename = str(tmp_path / self.filename)
        hdu1 = fits.PrimaryHDU()
        extname = "SCI"
        hdu2 = fits.ImageHDU(self.array)
        hdu2.header["EXTNAME"] = extname
        hdulist = fits.HDUList([hdu1, hdu2])
        hdulist.writeto(filename)
        main([filename, "-e", extname])

    @pytest.mark.parametrize("file_exten", [".gz", ".bz2"])
    def test_compressed_fits(self, tmp_path, file_exten):
        filename = str(tmp_path / f"test.fits{file_exten}")
        fits.writeto(filename, self.array)
        main([filename, "-e", "0"])

    def test_orientation(self, tmp_path):
        """
        Regression test to check the image vertical orientation/origin.
        """

        filename = str(tmp_path / self.filename)
        out_filename = "fits2bitmap_test.png"
        out_filename = str(tmp_path / out_filename)
        data = np.zeros((32, 32))
        data[0:16, :] = 1.0
        fits.writeto(filename, data)
        main([filename, "-e", "0", "-o", out_filename])

        img = mpimg.imread(out_filename)
        assert img[0, 0, 0] == 0
        assert img[31, 31, 0] == 1

    def test_min_max_cut_deprecations(self, tmp_path):
        filename = str(tmp_path / self.filename)
        fits.writeto(filename, self.array)
        with pytest.raises(SystemExit):
            main([filename, "--min_cut=0.1", "--vmin=0.1"])
        with pytest.raises(SystemExit):
            main([filename, "--max_cut=0.1", "--vmax=0.1"])



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import re
import warnings

from astropy.io.fits.column import KEYWORD_NAMES as TABLE_KEYWORD_NAMES
from astropy.io.fits.column import TDEF_RE, ColDefs, Column
from astropy.io.fits.hdu.compressed.compbintable import _CompBinTableHDU
from astropy.io.fits.header import Header
from astropy.io.fits.verify import VerifyWarning
from astropy.utils.exceptions import AstropyDeprecationWarning, AstropyUserWarning

from .settings import (
    CMTYPE_ALIASES,
    COMPRESSION_TYPES,
    DEFAULT_BLOCK_SIZE,
    DEFAULT_BYTE_PIX,
    DEFAULT_COMPRESSION_TYPE,
    DEFAULT_DITHER_SEED,
    DEFAULT_QUANTIZE_METHOD,
    NO_DITHER,
    QUANTIZE_METHOD_NAMES,
    SUBTRACTIVE_DITHER_1,
    SUBTRACTIVE_DITHER_2,
)
from .utils import _validate_tile_shape

__all__ = [
    "CompImageHeader",
    "_bintable_header_to_image_header",
    "_image_header_to_empty_bintable",
]

ZDEF_RE = re.compile(r"(?P<label>^[Zz][a-zA-Z]*)(?P<num>[1-9][0-9 ]*$)?")
INDEXED_COMPRESSION_KEYWORDS = {
    "ZNAXIS",
    "ZTILE",
    "ZNAME",
    "ZVAL",
}
REMAPPED_KEYWORDS = {
    "SIMPLE": "ZSIMPLE",
    "XTENSION": "ZTENSION",
    "BITPIX": "ZBITPIX",
    "NAXIS": "ZNAXIS",
    "EXTEND": "ZEXTEND",
    "BLOCKED": "ZBLOCKED",
    "PCOUNT": "ZPCOUNT",
    "GCOUNT": "ZGCOUNT",
    "CHECKSUM": "ZHECKSUM",
    "DATASUM": "ZDATASUM",
}
COMPRESSION_KEYWORDS = set(REMAPPED_KEYWORDS.values()).union(
    {
        "ZIMAGE",
        "ZCMPTYPE",
        "ZMASKCMP",
        "ZQUANTIZ",
        "ZDITHER0",
        "ZBLANK",
        "ZSCALE",
        "ZZERO",
    }
)


class CompImageHeader(Header):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "The CompImageHeader class is deprecated and will be removed in future",
            AstropyDeprecationWarning,
        )
        super().__init__(*args, **kwargs)


def _is_reserved_table_keyword(keyword):
    m = TDEF_RE.match(keyword)
    return keyword in ("TFIELDS", "THEAP") or (
        m and m.group("label").upper() in TABLE_KEYWORD_NAMES
    )


def _is_reserved_compression_keyword(keyword):
    m = ZDEF_RE.match(keyword)
    return keyword in COMPRESSION_KEYWORDS or (
        m and m.group("label").upper() in INDEXED_COMPRESSION_KEYWORDS
    )


def _is_reserved_keyword(keyword):
    return _is_reserved_table_keyword(keyword) or _is_reserved_compression_keyword(
        keyword
    )


def _bintable_header_to_image_header(bintable_header):
    # Start with a copy of the table header.
    image_header = bintable_header.copy()

    bscale = image_header.get("BSCALE")
    bzero = image_header.get("BZERO")

    # Strip out special keywords
    image_header.strip()

    # Delete cards that are related to the table.  And move
    # the values of those cards that relate to the image from
    # their corresponding table cards.  These include
    # ZBITPIX -> BITPIX, ZNAXIS -> NAXIS, and ZNAXISn -> NAXISn.
    # (Note: Used set here instead of list in case there are any duplicate
    # keywords, which there may be in some pathological cases:
    # https://github.com/astropy/astropy/issues/2750
    for keyword in set(image_header):
        if _is_reserved_keyword(keyword) or keyword in ("CHECKSUM", "DATASUM"):
            del image_header[keyword]

    if bscale:
        image_header["BSCALE"] = bscale
    if bzero:
        image_header["BZERO"] = bzero

    hcomments = bintable_header.comments

    if "ZSIMPLE" in bintable_header:
        image_header.set(
            "SIMPLE", bintable_header["ZSIMPLE"], hcomments["ZSIMPLE"], before=0
        )
    elif "ZTENSION" in bintable_header:
        if bintable_header["ZTENSION"] != "IMAGE":
            warnings.warn(
                "ZTENSION keyword in compressed extension != 'IMAGE'",
                AstropyUserWarning,
            )
        image_header.set("XTENSION", "IMAGE", hcomments["ZTENSION"], before=0)
    else:
        image_header.set("XTENSION", "IMAGE", before=0)

    image_header.set(
        "BITPIX", bintable_header["ZBITPIX"], hcomments["ZBITPIX"], before=1
    )

    image_header.set("NAXIS", bintable_header["ZNAXIS"], hcomments["ZNAXIS"], before=2)

    last_naxis = "NAXIS"
    for idx in range(image_header["NAXIS"]):
        znaxis = "ZNAXIS" + str(idx + 1)
        naxis = znaxis[1:]
        image_header.set(
            naxis, bintable_header[znaxis], hcomments[znaxis], after=last_naxis
        )
        last_naxis = naxis

    # Delete any other spurious NAXISn keywords:
    naxis = image_header["NAXIS"]
    for keyword in list(image_header["NAXIS?*"]):
        try:
            n = int(keyword[5:])
        except Exception:
            continue

        if n > naxis:
            del image_header[keyword]

    # Although PCOUNT and GCOUNT are considered mandatory for IMAGE HDUs,
    # ZPCOUNT and ZGCOUNT are optional, probably because for IMAGE HDUs
    # their values are always 0 and 1 respectively
    if "ZPCOUNT" in bintable_header:
        image_header.set(
            "PCOUNT",
            bintable_header["ZPCOUNT"],
            hcomments["ZPCOUNT"],
            after=last_naxis,
        )
    else:
        image_header.set("PCOUNT", 0, after=last_naxis)

    if "ZGCOUNT" in bintable_header:
        image_header.set(
            "GCOUNT", bintable_header["ZGCOUNT"], hcomments["ZGCOUNT"], after="PCOUNT"
        )
    else:
        image_header.set("GCOUNT", 1, after="PCOUNT")

    if "ZEXTEND" in bintable_header:
        image_header.set("EXTEND", bintable_header["ZEXTEND"], hcomments["ZEXTEND"])

    if "ZBLOCKED" in bintable_header:
        image_header.set("BLOCKED", bintable_header["ZBLOCKED"], hcomments["ZBLOCKED"])

    # Move the ZHECKSUM and ZDATASUM cards to the image header
    # as CHECKSUM and DATASUM
    if "ZHECKSUM" in bintable_header:
        image_header.set("CHECKSUM", bintable_header["ZHECKSUM"], hcomments["ZHECKSUM"])

    if "ZDATASUM" in bintable_header:
        image_header.set("DATASUM", bintable_header["ZDATASUM"], hcomments["ZDATASUM"])

    # Remove the EXTNAME card if the value in the table header
    # is the default value of COMPRESSED_IMAGE.
    if "EXTNAME" in image_header and image_header["EXTNAME"] == "COMPRESSED_IMAGE":
        del image_header["EXTNAME"]

    # Remove the PCOUNT GCOUNT cards if the uncompressed header is
    # from a primary HDU
    if "SIMPLE" in image_header:
        del image_header["PCOUNT"]
        del image_header["GCOUNT"]

    # Fill in BLANK keyword if necessary
    if image_header["BITPIX"] > 0 and "BLANK" not in image_header:
        if "ZBLANK" in bintable_header:
            image_header["BLANK"] = bintable_header["ZBLANK"]
        else:
            # check for column named "ZBLANK"
            for i in range(1, bintable_header["TFIELDS"] + 1):
                if bintable_header[f"TTYPE{i}"] == "ZBLANK":
                    # required BLANK keyword is missing
                    # use most negative value as default
                    image_header["BLANK"] = -(1 << (image_header["BITPIX"] - 1))
                    warnings.warn(
                        f"Setting default value {image_header['BLANK']} for "
                        "missing BLANK keyword in compressed extension",
                        AstropyUserWarning,
                    )
                    break

    # Look to see if there are any blank cards in the table
    # header.  If there are, there should be the same number
    # of blank cards in the image header.  Add blank cards to
    # the image header to make it so.
    table_blanks = bintable_header._countblanks()
    image_blanks = image_header._countblanks()

    for _ in range(table_blanks - image_blanks):
        image_header.append()

    return image_header


def _image_header_to_empty_bintable(
    image_header,
    name=None,
    huge_hdu=False,
    compression_type=None,
    tile_shape=None,
    hcomp_scale=None,
    hcomp_smooth=None,
    quantize_level=None,
    quantize_method=None,
    dither_seed=None,
    axes=None,
    generate_dither_seed=None,
):
    bintable = _CompBinTableHDU()

    # NOTE: image_header is the header that a user would see as the image
    # header which they might have set things like BSCALE and BZERO on, or
    # added history or comments to.

    # Update the extension name in the table header
    bintable.header.set(
        "EXTNAME",
        name,
        "name of this binary table extension",
        after="TFIELDS",
    )

    # Set the compression type in the table header.
    if compression_type:
        if compression_type not in COMPRESSION_TYPES:
            warnings.warn(
                "Unknown compression type provided (supported are {}). "
                "Default ({}) compression will be used.".format(
                    ", ".join(map(repr, COMPRESSION_TYPES)),
                    DEFAULT_COMPRESSION_TYPE,
                ),
                AstropyUserWarning,
            )
            compression_type = DEFAULT_COMPRESSION_TYPE

        bintable.header.set(
            "ZCMPTYPE", compression_type, "compression algorithm", after="TFIELDS"
        )
    else:
        compression_type = CMTYPE_ALIASES.get(compression_type, compression_type)

    # If the input image header had BSCALE/BZERO cards, then insert
    # them in the table header.

    bzero = image_header.get("BZERO", 0.0)
    bscale = image_header.get("BSCALE", 1.0)
    after_keyword = "EXTNAME"

    if bscale != 1.0:
        bintable.header.set("BSCALE", bscale, after=after_keyword)
        after_keyword = "BSCALE"

    if bzero != 0.0:
        bintable.header.set("BZERO", bzero, after=after_keyword)

    try:
        bitpix_comment = image_header.comments["BITPIX"]
    except (AttributeError, KeyError):
        bitpix_comment = "data type of original image"

    try:
        naxis_comment = image_header.comments["NAXIS"]
    except (AttributeError, KeyError):
        naxis_comment = "dimension of original image"

    # Set the label for the first column in the table

    bintable.header.set(
        "TTYPE1", "COMPRESSED_DATA", "label for field 1", after="TFIELDS"
    )

    # Set the data format for the first column.  It is dependent
    # on the requested compression type.

    if compression_type == "PLIO_1":
        tform1 = "1QI" if huge_hdu else "1PI"
    else:
        tform1 = "1QB" if huge_hdu else "1PB"

    bintable.header.set(
        "TFORM1",
        tform1,
        "data format of field: variable length array",
        after="TTYPE1",
    )

    # Create the first column for the table.  This column holds the
    # compressed data.
    col1 = Column(name=bintable.header["TTYPE1"], format=tform1)

    # Create the additional columns required for floating point
    # data and calculate the width of the output table.

    zbitpix = image_header["BITPIX"]

    if zbitpix < 0 and quantize_level != 0.0:
        # floating point image has 'COMPRESSED_DATA',
        # 'GZIP_COMPRESSED_DATA', 'ZSCALE', and 'ZZERO' columns (unless using
        # lossless compression, per CFITSIO)
        ncols = 4

        ttype2 = "GZIP_COMPRESSED_DATA"

        # The required format for the GZIP_COMPRESSED_DATA is actually
        # missing from the standard docs, but CFITSIO suggests it
        # should be 1PB, which is logical.
        tform2 = "1QB" if huge_hdu else "1PB"

        # Set up the second column for the table that will hold any
        # uncompressable data.
        bintable.header.set("TTYPE2", ttype2, "label for field 2", after="TFORM1")

        bintable.header.set(
            "TFORM2",
            tform2,
            "data format of field: variable length array",
            after="TTYPE2",
        )

        col2 = Column(name=ttype2, format=tform2)

        # Set up the third column for the table that will hold
        # the scale values for quantized data.
        bintable.header.set("TTYPE3", "ZSCALE", "label for field 3", after="TFORM2")
        bintable.header.set(
            "TFORM3", "1D", "data format of field: 8-byte DOUBLE", after="TTYPE3"
        )
        col3 = Column(name=bintable.header["TTYPE3"], format=bintable.header["TFORM3"])

        # Set up the fourth column for the table that will hold
        # the zero values for the quantized data.
        bintable.header.set("TTYPE4", "ZZERO", "label for field 4", after="TFORM3")
        bintable.header.set(
            "TFORM4", "1D", "data format of field: 8-byte DOUBLE", after="TTYPE4"
        )
        after = "TFORM4"
        col4 = Column(name=bintable.header["TTYPE4"], format=bintable.header["TFORM4"])

        # Create the ColDefs object for the table
        cols = ColDefs([col1, col2, col3, col4])
    else:
        # default table has just one 'COMPRESSED_DATA' column
        ncols = 1
        after = "TFORM1"

        # Create the ColDefs object for the table
        cols = ColDefs([col1])

    # Update the table header with the width of the table, the
    # number of fields in the table, the indicator for a compressed
    # image HDU, the data type of the image data and the number of
    # dimensions in the image data array.
    bintable.header.set("NAXIS1", cols.dtype.itemsize, "width of table in bytes")
    bintable.header.set(
        "TFIELDS", ncols, "number of fields in each row", after="GCOUNT"
    )
    bintable.header.set(
        "ZIMAGE", True, "extension contains compressed image", after=after
    )
    bintable.header.set("ZBITPIX", zbitpix, bitpix_comment, after="ZIMAGE")
    bintable.header.set("ZNAXIS", image_header["NAXIS"], naxis_comment, after="ZBITPIX")

    # Verify that any input tile size parameter is the appropriate
    # size to match the HDU's data.

    tile_shape = _validate_tile_shape(
        tile_shape=tile_shape,
        compression_type=compression_type,
        image_header=image_header,
    )

    # Set up locations for writing the next cards in the header.
    last_znaxis = "ZNAXIS"

    if image_header["NAXIS"] > 0:
        after1 = "ZNAXIS1"
    else:
        after1 = "ZNAXIS"

    # Calculate the number of rows in the output table and
    # write the ZNAXISn and ZTILEn cards to the table header.
    nrows = 0

    for idx, axis in enumerate(axes):
        naxis = "NAXIS" + str(idx + 1)
        znaxis = "ZNAXIS" + str(idx + 1)
        ztile = "ZTILE" + str(idx + 1)

        ts = tile_shape[len(axes) - 1 - idx]

        if not nrows:
            nrows = (axis - 1) // ts + 1
        else:
            nrows *= (axis - 1) // ts + 1

        if naxis in image_header:
            bintable.header.set(
                znaxis, axis, image_header.comments[naxis], after=last_znaxis
            )
        else:
            bintable.header.set(
                znaxis, axis, "length of original image axis", after=last_znaxis
            )

        bintable.header.set(ztile, ts, "size of tiles to be compressed", after=after1)
        last_znaxis = znaxis
        after1 = ztile

    # Set the NAXIS2 header card in the table hdu to the number of
    # rows in the table.
    bintable.header.set("NAXIS2", nrows, "number of rows in table")

    # Set the compression parameters in the table header based on the compression type

    after_keyword = "ZCMPTYPE"
    idx = 1

    if compression_type == "RICE_1":
        bintable.header.set(
            "ZNAME1", "BLOCKSIZE", "compression block size", after=after_keyword
        )
        bintable.header.set(
            "ZVAL1", DEFAULT_BLOCK_SIZE, "pixels per block", after="ZNAME1"
        )

        bintable.header.set(
            "ZNAME2", "BYTEPIX", "bytes per pixel (1, 2, 4, or 8)", after="ZVAL1"
        )

        if bintable.header["ZBITPIX"] == 8:
            bytepix = 1
        elif bintable.header["ZBITPIX"] == 16:
            bytepix = 2
        else:
            bytepix = DEFAULT_BYTE_PIX

        bintable.header.set(
            "ZVAL2", bytepix, "bytes per pixel (1, 2, 4, or 8)", after="ZNAME2"
        )
        after_keyword = "ZVAL2"
        idx = 3
    elif compression_type == "HCOMPRESS_1":
        bintable.header.set(
            "ZNAME1", "SCALE", "HCOMPRESS scale factor", after=after_keyword
        )
        bintable.header.set(
            "ZVAL1", hcomp_scale, "HCOMPRESS scale factor", after="ZNAME1"
        )
        bintable.header.set(
            "ZNAME2", "SMOOTH", "HCOMPRESS smooth option", after="ZVAL1"
        )
        bintable.header.set(
            "ZVAL2", hcomp_smooth, "HCOMPRESS smooth option", after="ZNAME2"
        )
        after_keyword = "ZVAL2"
        idx = 3

    if image_header["BITPIX"] < 0:  # floating point image
        bintable.header.set(
            "ZNAME" + str(idx),
            "NOISEBIT",
            "floating point quantization level",
            after=after_keyword,
        )
        bintable.header.set(
            "ZVAL" + str(idx),
            quantize_level,
            "floating point quantization level",
            after="ZNAME" + str(idx),
        )

        # Add the dither method and seed
        if quantize_method:
            if quantize_method not in [
                NO_DITHER,
                SUBTRACTIVE_DITHER_1,
                SUBTRACTIVE_DITHER_2,
            ]:
                name = QUANTIZE_METHOD_NAMES[DEFAULT_QUANTIZE_METHOD]
                warnings.warn(
                    "Unknown quantization method provided.  "
                    f"Default method ({name}) used."
                )
                quantize_method = DEFAULT_QUANTIZE_METHOD

            if quantize_method == NO_DITHER:
                zquantiz_comment = "No dithering during quantization"
            else:
                zquantiz_comment = "Pixel Quantization Algorithm"

            bintable.header.set(
                "ZQUANTIZ",
                QUANTIZE_METHOD_NAMES[quantize_method],
                zquantiz_comment,
                after="ZVAL" + str(idx),
            )
        else:
            # If the ZQUANTIZ keyword is missing the default is to assume
            # no dithering, rather than whatever DEFAULT_QUANTIZE_METHOD
            # is set to
            quantize_method = bintable.header.get("ZQUANTIZ", NO_DITHER)

            if isinstance(quantize_method, str):
                for k, v in QUANTIZE_METHOD_NAMES.items():
                    if v.upper() == quantize_method:
                        quantize_method = k
                        break
                else:
                    quantize_method = NO_DITHER

        if quantize_method == NO_DITHER:
            if "ZDITHER0" in bintable.header:
                # If dithering isn't being used then there's no reason to
                # keep the ZDITHER0 keyword
                del bintable.header["ZDITHER0"]
        else:
            if dither_seed:
                dither_seed = generate_dither_seed(dither_seed)
            elif "ZDITHER0" in bintable.header:
                dither_seed = bintable.header["ZDITHER0"]
            else:
                dither_seed = generate_dither_seed(DEFAULT_DITHER_SEED)

            bintable.header.set(
                "ZDITHER0",
                dither_seed,
                "dithering offset when quantizing floats",
                after="ZQUANTIZ",
            )

    # Move SIMPLE card from the image header to the
    # table header as ZSIMPLE card.

    if "SIMPLE" in image_header:
        bintable.header.set(
            "ZSIMPLE",
            image_header["SIMPLE"],
            image_header.comments["SIMPLE"],
            before="ZBITPIX",
        )

    # Move EXTEND card from the image header to the
    # table header as ZEXTEND card.

    if "EXTEND" in image_header:
        bintable.header.set(
            "ZEXTEND", image_header["EXTEND"], image_header.comments["EXTEND"]
        )

    # Move BLOCKED card from the image header to the
    # table header as ZBLOCKED card.

    if "BLOCKED" in image_header:
        bintable.header.set(
            "ZBLOCKED",
            image_header["BLOCKED"],
            image_header.comments["BLOCKED"],
        )

    # Move XTENSION card from the image header to the
    # table header as ZTENSION card.

    # Since we only handle compressed IMAGEs, ZTENSION should
    # always be IMAGE, even if the caller has passed in a header
    # for some other type of extension.
    if "XTENSION" in image_header:
        bintable.header.set(
            "ZTENSION",
            "IMAGE",
            image_header.comments["XTENSION"],
            before="ZBITPIX",
        )

    # Move PCOUNT and GCOUNT cards from image header to the table
    # header as ZPCOUNT and ZGCOUNT cards.

    if "PCOUNT" in image_header:
        bintable.header.set(
            "ZPCOUNT",
            image_header["PCOUNT"],
            image_header.comments["PCOUNT"],
            after=last_znaxis,
        )

    if "GCOUNT" in image_header:
        bintable.header.set(
            "ZGCOUNT",
            image_header["GCOUNT"],
            image_header.comments["GCOUNT"],
            after="ZPCOUNT",
        )

    # Move CHECKSUM and DATASUM cards from the image header to the
    # table header as XHECKSUM and XDATASUM cards.

    if "CHECKSUM" in image_header:
        bintable.header.set(
            "ZHECKSUM",
            image_header["CHECKSUM"],
            image_header.comments["CHECKSUM"],
        )

    if "DATASUM" in image_header:
        bintable.header.set(
            "ZDATASUM",
            image_header["DATASUM"],
            image_header.comments["DATASUM"],
        )

    bintable.columns = cols

    # Add any keywords that are in the original header that are not already
    # FIXME: don't use keyword_remaps, instead define an actual list to check
    # including regular expressions for NAXIS and other similar keywords

    # We need to add comments and history entries back to the right place, so
    # we first write out regular keywords and keep track of what the previous
    # regular keyword was for each position. We need to do this because
    # add_comment and add_history automatically write out to the bottom of the
    # header and there is no way to tell them to add them to the current
    # location.
    current_last_keyword = bintable.header.cards[-1].keyword
    previous_keyword = []

    for card in image_header.cards:
        # We deal with blanks in _writeto, and we handle COMMENT and HISTORY
        # only once we have written out the regular keywords. deal with after

        if card.keyword == "":
            bintable.header.add_blank()
        elif _is_reserved_keyword(card.keyword):
            warnings.warn(
                f"Keyword {card.keyword!r} is reserved "
                "for use by the FITS Tiled Image "
                "Convention so will be ignored",
                VerifyWarning,
            )
        elif (
            card.keyword not in ("", "COMMENT", "HISTORY")
            and card.keyword not in REMAPPED_KEYWORDS
            and card.keyword not in bintable.header
            and not card.keyword.startswith("NAXIS")
        ):
            bintable.header.append(card)
            current_last_keyword = card.keyword

        previous_keyword.append(current_last_keyword)

    # We loop over cards in reverse order for 'after=' to return the correct
    # order in the final header.
    for icard, card in list(enumerate(image_header.cards))[::-1]:
        if card.keyword == "COMMENT":
            bintable.header.add_comment(card.value, after=previous_keyword[icard])
        elif card.keyword == "HISTORY":
            bintable.header.add_history(card.value, after=previous_keyword[icard])

    # TODO: avoid writing the same comment multiple times

    return bintable



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from datetime import datetime

import pytest
from numpy.testing import assert_allclose, assert_equal

from astropy import units as u
from astropy.table import Column, Table
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time, TimeDelta
from astropy.timeseries.periodograms import BoxLeastSquares, LombScargle
from astropy.timeseries.sampled import TimeSeries
from astropy.units import Quantity, UnitsWarning
from astropy.utils.data import get_pkg_data_filename

INPUT_TIME = Time(["2016-03-22T12:30:31", "2015-01-21T12:30:32", "2016-03-22T12:30:40"])
PLAIN_TABLE = Table([[1, 2, 11], [3, 4, 1], [1, 1, 1]], names=["a", "b", "c"])

CSV_FILE = get_pkg_data_filename("data/sampled.csv")


def test_empty_initialization():
    ts = TimeSeries()
    ts["time"] = Time([50001, 50002, 50003], format="mjd")


def test_empty_initialization_invalid():
    # Make sure things crash when the first column added is not a time column
    ts = TimeSeries()
    with pytest.raises(
        ValueError,
        match=(
            r"TimeSeries object is invalid - expected 'time' as the first column but"
            r" found 'flux'"
        ),
    ):
        ts["flux"] = [1, 2, 3]


def test_initialize_only_time():
    ts = TimeSeries(time=INPUT_TIME)
    assert ts["time"] is ts.time
    # NOTE: the object in the table is a copy
    assert_equal(ts.time.isot, INPUT_TIME.isot)


def test_initialization_with_data():
    ts = TimeSeries(time=INPUT_TIME, data=[[10, 2, 3], [4, 5, 6]], names=["a", "b"])
    assert_equal(ts.time.isot, INPUT_TIME.isot)
    assert_equal(ts["a"], [10, 2, 3])
    assert_equal(ts["b"], [4, 5, 6])


def test_initialize_only_data():
    with pytest.raises(
        TypeError, match=r"Either 'time' or 'time_start' should be specified"
    ):
        TimeSeries(data=[[10, 2, 3], [4, 5, 6]], names=["a", "b"])


def test_initialization_with_table():
    ts = TimeSeries(time=INPUT_TIME, data=PLAIN_TABLE)
    assert ts.colnames == ["time", "a", "b", "c"]


def test_initialization_with_time_delta():
    ts = TimeSeries(
        time_start=datetime(2018, 7, 1, 10, 10, 10),
        time_delta=TimeDelta(3, format="sec"),
        data=[[10, 2, 3], [4, 5, 6]],
        names=["a", "b"],
    )
    assert_equal(
        ts.time.isot,
        [
            "2018-07-01T10:10:10.000",
            "2018-07-01T10:10:13.000",
            "2018-07-01T10:10:16.000",
        ],
    )


def test_initialization_missing_time_delta():
    with pytest.raises(
        TypeError, match=r"'time' is scalar, so 'time_delta' is required"
    ):
        TimeSeries(
            time_start=datetime(2018, 7, 1, 10, 10, 10),
            data=[[10, 2, 3], [4, 5, 6]],
            names=["a", "b"],
        )


def test_initialization_invalid_time_and_time_start():
    with pytest.raises(TypeError, match=r"Cannot specify both 'time' and 'time_start'"):
        TimeSeries(
            time=INPUT_TIME,
            time_start=datetime(2018, 7, 1, 10, 10, 10),
            data=[[10, 2, 3], [4, 5, 6]],
            names=["a", "b"],
        )


def test_initialization_invalid_time_delta():
    with pytest.raises(
        TypeError, match=r"'time_delta' should be a Quantity or a TimeDelta"
    ):
        TimeSeries(
            time_start=datetime(2018, 7, 1, 10, 10, 10),
            time_delta=[1, 4, 3],
            data=[[10, 2, 3], [4, 5, 6]],
            names=["a", "b"],
        )


def test_initialization_with_time_in_data():
    data = PLAIN_TABLE.copy()
    data["time"] = INPUT_TIME

    ts1 = TimeSeries(data=data)

    assert set(ts1.colnames) == {"time", "a", "b", "c"}
    assert all(ts1.time == INPUT_TIME)

    ts2 = TimeSeries(data=[[10, 2, 3], INPUT_TIME], names=["a", "time"])
    assert set(ts2.colnames) == {"time", "a"}
    assert all(ts2.time == INPUT_TIME)

    MESSAGE = r"'time' has been given both in the table and as a keyword argument"

    with pytest.raises(TypeError, match=MESSAGE):
        # Don't allow ambiguous cases of passing multiple 'time' columns
        TimeSeries(data=data, time=INPUT_TIME)

    with pytest.raises(TypeError, match=MESSAGE):
        # 'time' is a protected name, don't allow ambiguous cases
        TimeSeries(time=INPUT_TIME, data=[[10, 2, 3], INPUT_TIME], names=["a", "time"])


def test_initialization_n_samples():
    # Make sure things crash with incorrect n_samples

    with pytest.raises(
        TypeError,
        match=(
            r"'n_samples' has been given both and it is not the same length as the"
            r" input data."
        ),
    ):
        TimeSeries(time=INPUT_TIME, data=PLAIN_TABLE, n_samples=1000)


def test_initialization_length_mismatch():
    with pytest.raises(
        ValueError, match=r"Length of 'time' \(3\) should match data length \(2\)"
    ):
        TimeSeries(time=INPUT_TIME, data=[[10, 2], [4, 5]], names=["a", "b"])


def test_initialization_invalid_both_time_and_time_delta():
    with pytest.raises(
        TypeError,
        match=r"'time_delta' should not be specified since 'time' is an array",
    ):
        TimeSeries(time=INPUT_TIME, time_delta=TimeDelta(3, format="sec"))


def test_fold():
    times = Time([1, 2, 3, 8, 9, 12], format="unix")

    ts = TimeSeries(time=times)
    ts["flux"] = [1, 4, 4, 3, 2, 3]

    # Try without epoch time, as it should default to the first time and
    # wrapping at half the period.
    tsf = ts.fold(period=3.2 * u.s)
    assert isinstance(tsf.time, TimeDelta)
    assert_allclose(tsf.time.sec, [0, 1, -1.2, 0.6, -1.6, 1.4], rtol=1e-6)

    # Try with epoch time
    tsf = ts.fold(period=3.2 * u.s, epoch_time=Time(1.6, format="unix"))
    assert isinstance(tsf.time, TimeDelta)
    assert_allclose(tsf.time.sec, [-0.6, 0.4, 1.4, 0.0, 1.0, 0.8], rtol=1e-6, atol=1e-6)

    # Now with wrap_phase set to the full period
    tsf = ts.fold(period=3.2 * u.s, wrap_phase=3.2 * u.s)
    assert isinstance(tsf.time, TimeDelta)
    assert_allclose(tsf.time.sec, [0, 1, 2, 0.6, 1.6, 1.4], rtol=1e-6)

    # Now set epoch_phase to be 1/4 of the way through the phase
    tsf = ts.fold(period=3.2 * u.s, epoch_phase=0.8 * u.s)
    assert isinstance(tsf.time, TimeDelta)
    assert_allclose(tsf.time.sec, [0.8, -1.4, -0.4, 1.4, -0.8, -1.0], rtol=1e-6)

    # And combining epoch_phase and wrap_phase
    tsf = ts.fold(period=3.2 * u.s, epoch_phase=0.8 * u.s, wrap_phase=3.2 * u.s)
    assert isinstance(tsf.time, TimeDelta)
    assert_allclose(tsf.time.sec, [0.8, 1.8, 2.8, 1.4, 2.4, 2.2], rtol=1e-6)

    # Now repeat the above tests but with normalization applied

    # Try without epoch time, as it should default to the first time and
    # wrapping at half the period.
    tsf = ts.fold(period=3.2 * u.s, normalize_phase=True)
    assert isinstance(tsf.time, Quantity)
    assert_allclose(
        tsf.time.to_value(u.one),
        [0, 1 / 3.2, -1.2 / 3.2, 0.6 / 3.2, -1.6 / 3.2, 1.4 / 3.2],
        rtol=1e-6,
    )

    # Try with epoch time
    tsf = ts.fold(
        period=3.2 * u.s, epoch_time=Time(1.6, format="unix"), normalize_phase=True
    )
    assert isinstance(tsf.time, Quantity)
    assert_allclose(
        tsf.time.to_value(u.one),
        [-0.6 / 3.2, 0.4 / 3.2, 1.4 / 3.2, 0.0 / 3.2, 1.0 / 3.2, 0.8 / 3.2],
        rtol=1e-6,
        atol=1e-6,
    )

    # Now with wrap_phase set to the full period
    tsf = ts.fold(period=3.2 * u.s, wrap_phase=1, normalize_phase=True)
    assert isinstance(tsf.time, Quantity)
    assert_allclose(
        tsf.time.to_value(u.one),
        [0, 1 / 3.2, 2 / 3.2, 0.6 / 3.2, 1.6 / 3.2, 1.4 / 3.2],
        rtol=1e-6,
    )

    # Now set epoch_phase to be 1/4 of the way through the phase
    tsf = ts.fold(period=3.2 * u.s, epoch_phase=0.25, normalize_phase=True)
    assert isinstance(tsf.time, Quantity)
    assert_allclose(
        tsf.time.to_value(u.one),
        [0.8 / 3.2, -1.4 / 3.2, -0.4 / 3.2, 1.4 / 3.2, -0.8 / 3.2, -1.0 / 3.2],
        rtol=1e-6,
    )

    # And combining epoch_phase and wrap_phase
    tsf = ts.fold(
        period=3.2 * u.s, epoch_phase=0.25, wrap_phase=1, normalize_phase=True
    )
    assert isinstance(tsf.time, Quantity)
    assert_allclose(
        tsf.time.to_value(u.one),
        [0.8 / 3.2, 1.8 / 3.2, 2.8 / 3.2, 1.4 / 3.2, 2.4 / 3.2, 2.2 / 3.2],
        rtol=1e-6,
    )


def test_fold_invalid_options():
    times = Time([1, 2, 3, 8, 9, 12], format="unix")

    ts = TimeSeries(time=times)
    ts["flux"] = [1, 4, 4, 3, 2, 3]

    with pytest.raises(
        u.UnitsError, match="period should be a Quantity in units of time"
    ):
        ts.fold(period=3.2)

    with pytest.raises(
        u.UnitsError, match="period should be a Quantity in units of time"
    ):
        ts.fold(period=3.2 * u.m)

    with pytest.raises(
        u.UnitsError,
        match=(
            "epoch_phase should be a Quantity in units of "
            "time when normalize_phase=False"
        ),
    ):
        ts.fold(period=3.2 * u.s, epoch_phase=0.2)

    with pytest.raises(
        u.UnitsError,
        match=(
            "epoch_phase should be a dimensionless Quantity "
            "or a float when normalize_phase=True"
        ),
    ):
        ts.fold(period=3.2 * u.s, epoch_phase=0.2 * u.s, normalize_phase=True)

    with pytest.raises(
        u.UnitsError,
        match=(
            "wrap_phase should be a Quantity in units of "
            "time when normalize_phase=False"
        ),
    ):
        ts.fold(period=3.2 * u.s, wrap_phase=0.2)

    with pytest.raises(
        u.UnitsError,
        match="wrap_phase should be dimensionless when normalize_phase=True",
    ):
        ts.fold(period=3.2 * u.s, wrap_phase=0.2 * u.s, normalize_phase=True)

    with pytest.raises(
        ValueError, match="wrap_phase should be between 0 and the period"
    ):
        ts.fold(period=3.2 * u.s, wrap_phase=-0.1 * u.s)

    with pytest.raises(
        ValueError, match="wrap_phase should be between 0 and the period"
    ):
        ts.fold(period=3.2 * u.s, wrap_phase=-4.2 * u.s)

    with pytest.raises(ValueError, match="wrap_phase should be between 0 and 1"):
        ts.fold(period=3.2 * u.s, wrap_phase=-0.1, normalize_phase=True)

    with pytest.raises(ValueError, match="wrap_phase should be between 0 and 1"):
        ts.fold(period=3.2 * u.s, wrap_phase=2.2, normalize_phase=True)


def test_pandas():
    pandas = pytest.importorskip("pandas")

    df1 = pandas.DataFrame()
    df1["a"] = [1, 2, 3]
    df1.set_index(pandas.DatetimeIndex(INPUT_TIME.datetime64), inplace=True)

    ts = TimeSeries.from_pandas(df1)
    assert_equal(ts.time.isot, INPUT_TIME.isot)
    assert ts.colnames == ["time", "a"]
    assert len(ts.indices) == 1
    assert (ts.indices["time"].columns[0] == INPUT_TIME).all()

    ts_tcb = TimeSeries.from_pandas(df1, time_scale="tcb")
    assert ts_tcb.time.scale == "tcb"

    df2 = ts.to_pandas()
    assert (df2.index.values == pandas.Index(INPUT_TIME.datetime64).values).all()
    assert df2.columns == pandas.Index(["a"])
    assert (df1["a"] == df2["a"]).all()

    with pytest.raises(TypeError, match=r"Input should be a pandas DataFrame"):
        TimeSeries.from_pandas(None)

    df4 = pandas.DataFrame()
    df4["a"] = [1, 2, 3]

    with pytest.raises(TypeError, match=r"DataFrame does not have a DatetimeIndex"):
        TimeSeries.from_pandas(df4)


def test_read_time_missing():
    with pytest.raises(
        ValueError,
        match=(
            r"``time_column`` should be provided since the default Table readers are"
            r" being used\."
        ),
    ):
        TimeSeries.read(CSV_FILE, format="csv")


def test_read_time_wrong():
    with pytest.raises(
        ValueError, match=r"Time column 'abc' not found in the input data\."
    ):
        TimeSeries.read(CSV_FILE, time_column="abc", format="csv")


def test_read():
    timeseries = TimeSeries.read(CSV_FILE, time_column="Date", format="csv")
    assert timeseries.colnames == ["time", "A", "B", "C", "D", "E", "F", "G"]
    assert len(timeseries) == 11
    assert timeseries["time"].format == "iso"
    assert timeseries["A"].sum() == 266.5


@pytest.mark.remote_data(source="astropy")
def test_kepler_astropy():
    from astropy.units import UnitsWarning

    filename = get_pkg_data_filename("timeseries/kplr010666592-2009131110544_slc.fits")

    with pytest.warns(UnitsWarning):
        timeseries = TimeSeries.read(filename, format="kepler.fits")

    assert timeseries["time"].format == "isot"
    assert timeseries["time"].scale == "tdb"
    assert timeseries["sap_flux"].unit.to_string() == "electron / s"
    assert len(timeseries) == 14280
    assert len(timeseries.columns) == 20


@pytest.mark.remote_data(source="astropy")
def test_tess_astropy():
    filename = get_pkg_data_filename(
        "timeseries/hlsp_tess-data-alerts_tess_phot_00025155310-s01_tess_v1_lc.fits"
    )
    with pytest.warns((UserWarning, UnitsWarning)) as record:
        timeseries = TimeSeries.read(filename, format="tess.fits")
    # we might hit some warnings more than once, but the exact sequence probably
    # does not matter too much, so we'll just try to match the *set* of unique warnings
    unique_warnings = {(wm.category, wm.message.args[0]) for wm in record}
    assert len(record) != len(unique_warnings)
    expected = {
        (
            UnitsWarning,
            "'BJD - 2457000, days' did not parse as fits unit: "
            "At col 0, Unit 'BJD' not supported by the FITS standard.  "
            "If this is meant to be a custom unit, define it with 'u.def_unit'. "
            "To have it recognized inside a file reader or "
            "other code, enable it with 'u.add_enabled_units'. For details, see "
            "https://docs.astropy.org/en/latest/units/combining_and_defining.html",
        ),
        (
            UnitsWarning,
            "'e-/s' did not parse as fits unit: "
            "At col 0, Unit 'e' not supported by the FITS standard.  "
            "If this is meant to be a custom unit, define it with 'u.def_unit'. "
            "To have it recognized inside a file reader or other code, "
            "enable it with 'u.add_enabled_units'. For details, see "
            "https://docs.astropy.org/en/latest/units/combining_and_defining.html",
        ),
        (
            UnitsWarning,
            "'pixels' did not parse as fits unit: "
            "At col 0, Unit 'pixels' not supported by the FITS standard. "
            "Did you mean pixel? "
            "If this is meant to be a custom unit, define it with 'u.def_unit'. "
            "To have it recognized inside a file "
            "reader or other code, enable it with 'u.add_enabled_units'. For details, "
            "see https://docs.astropy.org/en/latest/units/combining_and_defining.html",
        ),
        (UserWarning, "Ignoring 815 rows with NaN times"),
    }
    assert unique_warnings == expected, (
        f"Got some unexpected warnings\n{unique_warnings - expected}"
    )

    assert timeseries["time"].format == "isot"
    assert timeseries["time"].scale == "tdb"
    assert timeseries["sap_flux"].unit.to_string() == "electron / s"
    assert len(timeseries) == 19261
    assert len(timeseries.columns) == 20


def test_required_columns():
    # Test the machinery that makes sure that the required columns are present

    ts = TimeSeries(time=INPUT_TIME, data=[[10, 2, 3], [4, 5, 6]], names=["a", "b"])

    # In the examples below, the operation (e.g. remove_column) is actually
    # carried out before the checks are made, so we need to use copy() so that
    # we don't change the main version of the time series.

    # Make sure copy works fine
    ts.copy()

    MESSAGE = (
        r"TimeSeries object is invalid - expected 'time' as the first column but found"
        r" '{}'"
    )

    with pytest.raises(ValueError, match=MESSAGE.format("c")):
        ts.copy().add_column(Column([3, 4, 5], name="c"), index=0)

    with pytest.raises(ValueError, match=MESSAGE.format("d")):
        ts.copy().add_columns(
            [Column([3, 4, 5], name="d"), Column([3, 4, 5], name="e")], indexes=[0, 1]
        )

    with pytest.raises(ValueError, match=MESSAGE.format("a")):
        ts.copy().keep_columns(["a", "b"])

    with pytest.raises(ValueError, match=MESSAGE.format("a")):
        ts.copy().remove_column("time")

    with pytest.raises(ValueError, match=MESSAGE.format("b")):
        ts.copy().remove_columns(["time", "a"])

    with pytest.raises(ValueError, match=MESSAGE.format("banana")):
        ts.copy().rename_column("time", "banana")

    # https://github.com/astropy/astropy/issues/13009
    MESSAGE = (
        r"TimeSeries object is invalid - expected \['time', 'a'\] as the first columns"
        r" but found \['time', 'b'\]"
    )
    ts_2cols_required = ts.copy()
    ts_2cols_required._required_columns = ["time", "a"]
    with pytest.raises(ValueError, match=MESSAGE):
        ts_2cols_required.remove_column("a")


@pytest.mark.parametrize("cls", [BoxLeastSquares, LombScargle])
def test_periodogram(cls):
    # Note that we don't need to check the actual results from the periodogram
    # classes here since these are tested extensively in
    # astropy.timeseries.periodograms.

    ts = TimeSeries(time=INPUT_TIME, data=[[10, 2, 3], [4, 5, 6]], names=["a", "b"])

    p1 = cls.from_timeseries(ts, "a")
    assert isinstance(p1, cls)
    assert_allclose(p1.t.jd, ts.time.jd)
    assert_equal(p1.y, ts["a"])
    assert p1.dy is None

    p2 = cls.from_timeseries(ts, "a", uncertainty="b")
    assert_quantity_allclose(p2.dy, ts["b"])

    p3 = cls.from_timeseries(ts, "a", uncertainty=0.1)
    assert_allclose(p3.dy, 0.1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from astropy.io.votable import ucd


def test_none():
    assert ucd.check_ucd(None)


examples = {
    "phys.temperature": [("ivoa", "phys.temperature")],
    "pos.eq.ra;meta.main": [("ivoa", "pos.eq.ra"), ("ivoa", "meta.main")],
    "meta.id;src": [("ivoa", "meta.id"), ("ivoa", "src")],
    "phot.flux;em.radio;arith.ratio": [
        ("ivoa", "phot.flux"),
        ("ivoa", "em.radio"),
        ("ivoa", "arith.ratio"),
    ],
    "PHot.Flux;EM.Radio;ivoa:arith.Ratio": [
        ("ivoa", "phot.flux"),
        ("ivoa", "em.radio"),
        ("ivoa", "arith.ratio"),
    ],
    "pos.galactic.lat": [("ivoa", "pos.galactic.lat")],
    "meta.code;phot.mag": [("ivoa", "meta.code"), ("ivoa", "phot.mag")],
    "stat.error;phot.mag": [("ivoa", "stat.error"), ("ivoa", "phot.mag")],
    "phys.temperature;instr;stat.max": [
        ("ivoa", "phys.temperature"),
        ("ivoa", "instr"),
        ("ivoa", "stat.max"),
    ],
    "stat.error;phot.mag;em.opt.V": [
        ("ivoa", "stat.error"),
        ("ivoa", "phot.mag"),
        ("ivoa", "em.opt.V"),
    ],
    "phot.color;em.opt.B;em.opt.V": [
        ("ivoa", "phot.color"),
        ("ivoa", "em.opt.B"),
        ("ivoa", "em.opt.V"),
    ],
    "stat.error;phot.color;em.opt.B;em.opt.V": [
        ("ivoa", "stat.error"),
        ("ivoa", "phot.color"),
        ("ivoa", "em.opt.B"),
        ("ivoa", "em.opt.V"),
    ],
}


def test_check():
    for s, p in examples.items():
        assert ucd.parse_ucd(s, True, True) == p
        assert ucd.check_ucd(s, True, True)


def test_too_many_colons():
    with pytest.raises(ValueError):
        ucd.parse_ucd("ivoa:stsci:phot", True, True)


def test_invalid_namespace():
    with pytest.raises(ValueError):
        ucd.parse_ucd("_ivoa:phot.mag", True, True)


def test_invalid_word():
    with pytest.raises(ValueError):
        ucd.parse_ucd("-pho")


def test_atmospheric_wind_ucd():
    """Test parsing of atmospheric observation UCDs, specifically obs.atmos.wind.

    This test addresses issue #18452 where obs.atmos.wind and other new
    atmospheric terms from UCD1+ version 1.6 should be recognized.
    """
    # Test the specific case from the issue report
    ucd_string = "phys.veloc;obs.atmos.wind;stat.mean"
    result = ucd.parse_ucd(ucd_string, check_controlled_vocabulary=True)
    expected = [
        ("ivoa", "phys.veloc"),
        ("ivoa", "obs.atmos.wind"),
        ("ivoa", "stat.mean"),
    ]
    assert result == expected

    # Test primary/quantified atmospheric terms (can be used standalone)
    primary_atmos_terms = [
        "obs.atmos.humidity",
        "obs.atmos.rain",
        "obs.atmos.turbulence.isoplanatic",
        "obs.atmos.water",
        "phys.temperature.dew",
    ]

    for term in primary_atmos_terms:
        result = ucd.parse_ucd(term, check_controlled_vocabulary=True)
        assert result == [("ivoa", term)]
        assert ucd.check_ucd(term, check_controlled_vocabulary=True)

    # Test secondary atmospheric terms (must be used with primary terms)
    secondary_atmos_terms = [
        "obs.atmos.wind",
        "obs.atmos.turbulence",
    ]

    for term in secondary_atmos_terms:
        # Test as secondary word (valid)
        combined_ucd = f"phys.temperature;{term}"
        result = ucd.parse_ucd(combined_ucd, check_controlled_vocabulary=True)
        expected = [("ivoa", "phys.temperature"), ("ivoa", term)]
        assert result == expected
        assert ucd.check_ucd(combined_ucd, check_controlled_vocabulary=True)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import astropy.coordinates.representation as r
import astropy.units as u
from astropy.coordinates import FK5, SkyCoord
from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.coordinates.tests.helper import skycoord_equal
from astropy.coordinates.tests.test_representation import representation_equal
from astropy.utils.masked import Masked


class MaskedSphericalSetup:
    @classmethod
    def setup_class(cls):
        cls.lon = np.array([0.0, 3.0, 6.0, 12.0, 15.0, 18.0]) << u.hourangle
        cls.lat = np.array([-15.0, 30.0, 60.0, -60.0, 89.0, -80.0]) << u.deg
        cls.dis = np.array([10.0, 20.0, 30.0, 40.0, 50.0, 60.0]) << u.pc
        cls.mask_lon = np.array([False, True, False, False, True, True])
        cls.mask_lat = np.array([False, False, True, False, True, True])
        cls.mask_dis = np.array([False, False, False, True, False, True])
        cls.mlon = Masked(cls.lon, cls.mask_lon)
        cls.mlat = Masked(cls.lat, cls.mask_lat)
        cls.mdis = Masked(cls.dis, cls.mask_dis)
        cls.msph = r.SphericalRepresentation(cls.mlon, cls.mlat, cls.mdis)
        cls.mask_ang = cls.mask_lon | cls.mask_lat
        cls.mask = cls.mask_ang | cls.mask_dis


class TestSphericalRepresentationSeparateMasks(MaskedSphericalSetup):
    """Tests for mask propagation for Spherical with separate masks."""

    def test_initialization(self):
        assert_array_equal(self.msph.lon.mask, self.mask_lon)
        assert_array_equal(self.msph.lat.mask, self.mask_lat)
        assert_array_equal(self.msph.distance.mask, self.mask_dis)
        assert_array_equal(self.msph.unmasked.lon, self.lon)
        assert_array_equal(self.msph.unmasked.lat, self.lat)
        assert_array_equal(self.msph.unmasked.distance, self.dis)

        assert_array_equal(self.msph.mask, self.mask)
        assert_array_equal(self.msph.get_mask(), self.mask)
        assert_array_equal(self.msph.get_mask("lon", "lat"), self.mask_ang)
        assert repr(self.msph) == (
            "<SphericalRepresentation (lon, lat, distance) in (hourangle, deg, pc)\n"
            "    [( 0., -15., 10.), (———,  30., 20.), ( 6.,  ———, 30.),\n"
            "     (12., -60., ———), (———,  ———, 50.), (———,  ———, ———)]>"
        )
        assert str(self.msph) == (
            "[( 0., -15., 10.), (———,  30., 20.), ( 6.,  ———, 30.), (12., -60., ———),\n"
            " (———,  ———, 50.), (———,  ———, ———)] (hourangle, deg, pc)"
        )

    def test_convert_to_cartesian(self):
        mcart = self.msph.represent_as(r.CartesianRepresentation)
        assert_array_equal(mcart.mask, self.mask)

    def test_convert_to_unit_spherical(self):
        musph = self.msph.represent_as(r.UnitSphericalRepresentation)
        assert_array_equal(musph.lon.mask, self.mask_lon)
        assert_array_equal(musph.lat.mask, self.mask_lat)
        assert_array_equal(musph.mask, self.mask_ang)

    def test_convert_to_radial(self):
        mrad = r.RadialRepresentation.from_representation(self.msph)
        assert_array_equal(mrad.mask, self.mask_dis)

    def test_convert_to_physics_spherical(self):
        mpsph = self.msph.represent_as(r.PhysicsSphericalRepresentation)
        assert_array_equal(mpsph.phi.mask, self.mask_lon)
        assert_array_equal(mpsph.theta.mask, self.mask_lat)
        assert_array_equal(mpsph.r.mask, self.mask_dis)
        assert_array_equal(mpsph.mask, self.mask)
        assert_array_equal(mpsph.get_mask("phi", "theta"), self.mask_ang)

    def test_set_mask(self):
        msph = self.msph.copy()
        msph[0] = np.ma.masked
        assert_array_equal(msph.mask, np.concatenate(([True], self.mask[1:])))
        msph[0] = np.ma.nomask
        assert_array_equal(msph.mask, self.mask)

    def test_setitem(self):
        msph = self.msph.copy()
        msph[0] = self.msph[1]
        assert_array_equal(msph.mask, np.concatenate(([True], self.mask[1:])))
        assert_array_equal(
            msph.unmasked.lon, np.concatenate((msph.lon[1:2], self.lon[1:]))
        )
        msph[0] = self.msph[0]
        assert_array_equal(msph.mask, self.mask)
        assert_array_equal(msph.unmasked.lon, self.lon)

    def test_set_masked_item_on_unmasked_instance(self):
        # Currently, the mask on items is *ignored*, just as it is for ndarray,
        # Quantity, and Time. In principle, this could be changed for containers
        # like representations and Time. See
        # https://github.com/astropy/astropy/pull/17016#issuecomment-2439607869
        sph = self.msph.unmasked.copy()
        sph[0] = self.msph[1]
        assert not sph.masked
        assert_array_equal(sph.mask, np.zeros_like(self.mask))
        assert_array_equal(sph.lon, np.concatenate((sph.lon[1:2], self.lon[1:])))
        sph[0] = self.msph[0].unmasked
        assert not sph.masked
        assert_array_equal(sph.mask, np.zeros_like(self.mask))
        assert_array_equal(sph.unmasked.lon, self.lon)

    def test_set_np_ma_masked_on_unmasked_instance(self):
        sph = self.msph.copy().unmasked
        sph[0] = np.ma.masked
        assert sph.masked
        assert_array_equal(
            sph.mask, np.concatenate(([True], np.zeros_like(self.mask[1:])))
        )
        assert_array_equal(sph.unmasked.lon, self.lon)
        sph[0] = np.ma.nomask
        assert sph.masked  # Does not get reset
        assert_array_equal(sph.mask, np.zeros_like(self.mask))
        assert_array_equal(sph.unmasked.lon, self.lon)

    def test_set_np_ma_nomasked_on_unmasked_instance(self):
        sph = self.msph.copy().unmasked
        sph[0] = np.ma.nomask
        assert not sph.masked
        assert_array_equal(sph.unmasked.lon, self.lon)

    def test_filled(self):
        unmasked = self.msph.unmasked
        sph = self.msph.filled(unmasked[1])
        expected = unmasked.copy()
        expected[self.mask] = unmasked[1]
        assert np.all(representation_equal(sph, expected))

    def test_filled_with_masked_value(self):
        # Filled ignores the mask (this will be true as long as __setitem__
        # ignores it; it may be a logical choice to actually use it).
        sph = self.msph.filled(self.msph[1])
        assert not sph.masked
        expected = self.msph.unmasked.copy()
        expected[self.mask] = self.msph.unmasked[1]
        assert np.all(representation_equal(sph, expected))

    def test_transform_keeps_distance_angular_masks(self):
        m = rotation_matrix(30.0, "x") * 2.0  # rotation and scale
        sph = self.msph.transform(m)
        assert sph.masked
        # Distance now also masked if angular coordinates were masked.
        assert_array_equal(sph.distance.mask, self.mask)
        # But angular coordinates just depend on the angular mask.
        assert_array_equal(sph.lon.mask, self.mask_ang)
        assert_array_equal(sph.lat.mask, self.mask_ang)
        assert_array_equal(self.msph.get_mask(), self.mask)
        assert_array_equal(self.msph.get_mask("lon", "lat"), self.mask_ang)

    def test_unmasked_representation_masked_differential(self):
        rv = np.arange(6.0) << u.km / u.s
        mask_rv = [True, False] * 3
        mrv = Masked(rv, mask_rv)
        mdiff = r.RadialDifferential(mrv)
        msph = r.SphericalRepresentation(
            self.lon,
            self.lat,
            self.dis,
            differentials={"s": mdiff},
        )
        assert msph.masked
        assert_array_equal(msph.lon.mask, False)
        assert_array_equal(msph.lat.mask, False)
        assert_array_equal(msph.distance.mask, False)
        assert_array_equal(msph.differentials["s"].d_distance.mask, mask_rv)
        sph = msph.unmasked
        assert not sph.masked
        assert not sph.differentials["s"].masked
        # Sanity checks on "with[out]_differentials"
        assert msph.without_differentials().masked
        sph2 = r.SphericalRepresentation(self.lon, self.lat, self.dis)
        assert not sph2.masked
        sph3 = sph2.with_differentials({"s": mdiff})
        assert sph3.masked

    def test_masked_representation_unmasked_differential(self):
        diff = r.RadialDifferential(np.arange(6.0) << u.km / u.s)
        msph = r.SphericalRepresentation(
            self.mlon,
            self.mlat,
            self.mdis,
            differentials={"s": diff},
        )
        assert msph.masked
        assert msph.differentials["s"].masked
        assert_array_equal(msph.differentials["s"].d_distance.mask, False)
        # Sanity check on using with_differentials.
        msph2 = self.msph.with_differentials({"s": diff})
        assert msph2.masked
        assert msph2.differentials["s"].masked


class TestFrame(MaskedSphericalSetup):
    """Tests that mask is calculated properly for frames, using FK5."""

    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.fk5 = FK5(cls.msph)

    def test_initialization_directly(self):
        d = Masked([50, 1.0] * u.kpc, mask=[False, True])
        fk5 = FK5([0, 30] * u.deg, [-10, 10] * u.deg, distance=d)
        assert fk5.masked
        assert_array_equal(fk5.ra.mask, [False, False])
        assert_array_equal(fk5.dec.mask, [False, False])
        assert_array_equal(fk5.distance.mask, [False, True])
        assert_array_equal(fk5.mask, [False, True])
        assert_array_equal(fk5.get_mask("ra", "dec"), False)
        assert "—" in repr(fk5)

    def test_class_initialization(self):
        assert_array_equal(self.fk5.ra.mask, self.mask_lon)
        assert_array_equal(self.fk5.dec.mask, self.mask_lat)
        assert_array_equal(self.fk5.distance.mask, self.mask_dis)

        assert_array_equal(self.fk5.mask, self.mask)
        assert_array_equal(self.fk5.get_mask("ra", "dec"), self.mask_ang)

        unmasked = self.fk5.unmasked
        assert_array_equal(unmasked.ra, self.lon)
        assert_array_equal(unmasked.dec, self.lat)
        assert_array_equal(unmasked.distance, self.dis)

    def test_cache_clearing(self):
        mfk5 = self.fk5.copy()
        assert_array_equal(mfk5.ra.mask, self.mask_lon)
        assert "—" in repr(mfk5)
        mfk5[...] = np.ma.nomask
        assert_array_equal(mfk5.data.mask, np.zeros(mfk5.shape, bool))
        assert_array_equal(mfk5.ra.mask, np.zeros(mfk5.shape, bool))
        assert "—" not in repr(mfk5)
        mfk5[...] = self.fk5
        assert_array_equal(mfk5.ra.mask, self.mask_lon)
        assert "—" in repr(mfk5)

    def test_cartesian(self):
        mcart = FK5(self.msph.represent_as(r.CartesianRepresentation))
        assert_array_equal(mcart.mask, self.mask)

    def test_unit_spherical(self):
        musph = FK5(self.msph.represent_as(r.UnitSphericalRepresentation))
        assert_array_equal(musph.mask, self.mask_ang)
        assert_array_equal(musph.get_mask(), self.mask_ang)

    def test_physics_spherical(self):
        mpsph = FK5(self.msph.represent_as(r.PhysicsSphericalRepresentation))
        assert_array_equal(mpsph.mask, self.mask)
        assert_array_equal(mpsph.get_mask("data.phi", "data.theta"), self.mask_ang)

    def test_get_mask(self):
        assert_array_equal(self.fk5.get_mask("ra"), self.mask_lon)
        assert_array_equal(self.fk5.get_mask("ra", "dec"), self.mask_ang)
        assert_array_equal(self.fk5.get_mask("data.lat"), self.mask_lat)
        assert_array_equal(self.fk5.get_mask("data.lat", "data.lon"), self.mask_ang)
        assert_array_equal(self.fk5.get_mask("cartesian"), self.mask)
        assert_array_equal(self.fk5.get_mask("equinox"), np.zeros(self.fk5.shape, bool))

    def test_unmasked_frame(self):
        fk5 = self.fk5.unmasked
        assert not fk5.masked
        assert_array_equal(fk5.mask, np.zeros(fk5.shape, bool))
        assert_array_equal(fk5.get_mask(), np.zeros(fk5.shape, bool))


def test_frame_without_data():
    fk5_no_data = FK5(equinox=["J2000", "J2001"])
    with pytest.raises(ValueError, match="does not have associated data"):
        fk5_no_data.masked
    with pytest.raises(ValueError, match="does not have associated data"):
        fk5_no_data.mask
    with pytest.raises(ValueError, match="does not have associated data"):
        fk5_no_data.get_mask()
    assert_array_equal(fk5_no_data.get_mask("equinox"), np.zeros((2,), bool))


class TestSkyCoord(TestFrame):
    """Tests that mask is calculated properly for SkyCoord.

    Note that this does all the tests from TestFrame, as well as a few
    specific to SkyCoord, i.e., that use attributes the frame does not have.
    """

    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.p = np.linspace(900, 1000, cls.msph.size) << u.hPa
        cls.mask_p = np.array([True, False, False, False, False, False])
        cls.mp = Masked(cls.p, cls.mask_p)
        # Ensure we have an attribute not associated with the frame.
        cls.fk5 = SkyCoord(cls.msph, frame="fk5", pressure=cls.mp)

    def test_non_frame_attribute(self):
        assert_array_equal(self.fk5.get_mask("pressure"), self.mask_p)
        assert_array_equal(
            self.fk5.get_mask("pressure", "data"), self.mask | self.mask_p
        )


class TestSkyCoordWithDifferentials:
    @classmethod
    def setup_class(cls):
        cls.ra = [0.0, 3.0, 6.0, 12.0, 15.0, 18.0] << u.hourangle
        cls.dec = [-15.0, 30.0, 60.0, -60.0, 89.0, -80.0] << u.deg
        cls.dis = [10.0, 20.0, 30.0, 40.0, 50.0, 60.0] << u.pc
        cls.mask_dis = np.array([False, True, False, True, False, True])
        cls.pm_ra_cosdec = [1.0, 2.0, 3.0, -4.0, -5.0, -6.0] << (u.mas / u.yr)
        cls.mask_pm_ra_cosdec = np.array([False, False, True, False, False, True])
        cls.pm_dec = [-9.0, -7.0, 5.0, 3.0, 1.0, 0.0] << (u.mas / u.yr)
        cls.mask_pm_dec = np.array([False, False, True, True, False, True])
        cls.rv = [40.0, 50.0, 0.0, 0.0, -30.0, -10.0] << (u.km / u.s)
        cls.mask_rv = np.array([False, False, False, False, True, True])
        cls.mdis = Masked(cls.dis, cls.mask_dis)
        cls.mpm_ra_cosdec = Masked(cls.pm_ra_cosdec, cls.mask_pm_ra_cosdec)
        cls.mpm_dec = Masked(cls.pm_dec, cls.mask_pm_dec)
        cls.mrv = Masked(cls.rv, cls.mask_rv)
        cls.sc = SkyCoord(
            ra=cls.ra,
            dec=cls.dec,
            distance=cls.mdis,
            pm_ra_cosdec=cls.mpm_ra_cosdec,
            pm_dec=cls.mpm_dec,
            radial_velocity=cls.mrv,
        )
        cls.mask = cls.mask_dis | cls.mask_pm_ra_cosdec | cls.mask_pm_dec | cls.mask_rv

    def test_setup(self):
        assert self.sc.masked
        assert_array_equal(self.sc.ra.mask, False)
        assert_array_equal(self.sc.dec.mask, False)
        assert_array_equal(self.sc.distance.mask, self.mask_dis)
        assert_array_equal(self.sc.pm_ra_cosdec.mask, self.mask_pm_ra_cosdec)
        assert_array_equal(self.sc.pm_dec.mask, self.mask_pm_dec)
        assert_array_equal(self.sc.radial_velocity.mask, self.mask_rv)
        assert_array_equal(self.sc.mask, self.mask)

    def test_get_mask(self):
        assert_array_equal(self.sc.get_mask(), self.mask)
        assert_array_equal(self.sc.get_mask("ra", "dec"), False)

    def test_filled(self):
        unmasked = self.sc.unmasked
        filled = self.sc.filled(unmasked[1])
        expected = unmasked.copy()
        expected[self.mask] = unmasked[1]
        assert skycoord_equal(filled, expected)

    def test_filled_with_masked_value(self):
        # Filled ignores the mask (this will be true as long as __setitem__
        # ignores it; it may be a logical choice to actually use it).
        filled = self.sc.filled(self.sc[1])
        expected = self.sc.unmasked.copy()
        expected[self.mask] = self.sc.unmasked[1]
        assert skycoord_equal(filled, expected)

    @pytest.mark.parametrize(
        "dt",
        [
            1 * u.yr,
            Masked([1, 2, 3] * u.yr, mask=[False, True, False])[:, np.newaxis],
        ],
    )
    def test_apply_space_motion(self, dt):
        sc = self.sc.apply_space_motion(dt=dt)
        # All parts of the coordinate influence the final positions.
        expected_mask = self.sc.get_mask() | getattr(dt, "mask", False)
        assert_array_equal(sc.mask, expected_mask)
        expected_unmasked = self.sc.unmasked.apply_space_motion(dt=dt)
        assert skycoord_equal(sc.unmasked, expected_unmasked)


class TestSkyCoordWithOnlyDifferentialsMasked(TestSkyCoordWithDifferentials):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        # Overwrite SkyCoord using unmasked distance.
        cls.mask_dis = False
        cls.sc = SkyCoord(
            ra=cls.ra,
            dec=cls.dec,
            distance=cls.dis,
            pm_ra_cosdec=cls.mpm_ra_cosdec,
            pm_dec=cls.mpm_dec,
            radial_velocity=cls.mrv,
        )
        cls.mask = cls.mask_dis | cls.mask_pm_ra_cosdec | cls.mask_pm_dec | cls.mask_rv



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Script support for validating a VO file.
"""

import argparse


def main(args=None):
    from . import table

    parser = argparse.ArgumentParser(
        description="Check a VOTable file for compliance to the VOTable specification"
    )
    # TODO: pass suggest_on_error as kwarg when PYTHON_LT_14 is dropped
    parser.suggest_on_error = True

    parser.add_argument("filename", nargs=1, help="Path to VOTable file to check")
    args = parser.parse_args(args)

    table.validate(args.filename[0])



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from itertools import pairwise

import numpy as np

from astropy import units as u
from astropy.time import Time, TimeDelta
from astropy.timeseries.binned import BinnedTimeSeries
from astropy.timeseries.sampled import TimeSeries
from astropy.utils.exceptions import AstropyUserWarning
from astropy.utils.masked import Masked, get_data_and_mask

__all__ = ["aggregate_downsample"]


def nanmean_reduceat(data, indices):
    """Calculate nanmean on pieces of data given by the reduceat-like indices.

    Will skip both masked elements and those equal to nan.

    The output for any fully-masked pieces is set to nan (and will be masked
    if the input was masked).
    """
    unmasked, data_mask = get_data_and_mask(data)
    mask = np.isnan(unmasked)
    if data_mask is not None:
        mask |= data_mask

    if mask.any():  # If there are NaNs or masked elements
        # Create a writeable copy and set masked/NaN elements to zero.
        unmasked = unmasked.copy()
        unmasked[mask] = 0
        counts = np.add.reduceat(~mask, indices)
        out_mask = counts <= 0
    else:
        # Derive counts from indices
        counts = np.diff(indices, append=len(data))
        out_mask = None

    result = np.add.reduceat(unmasked, indices) / np.maximum(counts, 1)
    if out_mask is not None:
        result[out_mask] = np.nan
    if data_mask is not None:  # Had masked input.
        # use of Masked is needed for case data is astropy.utils.masked.core.MaskedNDArray,
        # to avoid a TypeError from MaskedNDArray
        cls = Masked if isinstance(data, Masked) else data.__class__
        result = cls(result, mask=out_mask, copy=False)
    return result


def reduceat(array, indices, function):
    """
    Manual reduceat functionality for cases where Numpy functions don't have a reduceat.
    It will check if the input function has a reduceat and call that if it does.
    """
    if len(indices) == 0:
        return np.zeros_like(array, shape=(0,))
    elif function is nanmean_reduceat:
        return function(array, indices)
    elif hasattr(function, "reduceat"):
        return function.reduceat(array, indices)
    else:
        result = [
            function(array[start:stop] if start < stop else array[start])
            for start, stop in pairwise(list(indices) + [len(array)])
        ]
        return np.block(result)


def _to_relative_longdouble(time: Time, rel_base: Time) -> np.longdouble:
    # Convert the time objects into plain ndarray
    # so that they be used to make various operations faster, including
    # - `np.searchsorted()`
    # - time comparison.
    #
    # Relative time in seconds with np.longdouble type is used to:
    # - a consistent format for search, irrespective of the format/scale of the inputs,
    # - retain the best precision possible
    return (time - rel_base).to_value(format="sec", subfmt="long")


def aggregate_downsample(
    time_series,
    *,
    time_bin_size=None,
    time_bin_start=None,
    time_bin_end=None,
    n_bins=None,
    aggregate_func=None,
):
    """
    Downsample a time series by binning values into bins with a fixed size or
    custom sizes, using a single function to combine the values in the bin.

    Parameters
    ----------
    time_series : :class:`~astropy.timeseries.TimeSeries`
        The time series to downsample.
    time_bin_size : `~astropy.units.Quantity` or `~astropy.time.TimeDelta` ['time'], optional
        The time interval for the binned time series - this is either a scalar
        value (in which case all time bins will be assumed to have the same
        duration) or as an array of values (in which case each time bin can
        have a different duration). If this argument is provided,
        ``time_bin_end`` should not be provided.
    time_bin_start : `~astropy.time.Time` or iterable, optional
        The start time for the binned time series - this can be either given
        directly as a `~astropy.time.Time` array or as any iterable that
        initializes the `~astropy.time.Time` class. This can also be a scalar
        value if ``time_bin_size`` or ``time_bin_end`` is provided.
        Defaults to the first time in the sampled time series.
    time_bin_end : `~astropy.time.Time` or iterable, optional
        The times of the end of each bin - this can be either given directly as
        a `~astropy.time.Time` array or as any iterable that initializes the
        `~astropy.time.Time` class. This can only be given if ``time_bin_start``
        is provided or its default is used. If ``time_bin_end`` is scalar and
        ``time_bin_start`` is an array, time bins are assumed to be contiguous;
        the end of each bin is the start of the next one, and ``time_bin_end`` gives
        the end time for the last bin.  If ``time_bin_end`` is an array and
        ``time_bin_start`` is scalar, bins will be contiguous. If both ``time_bin_end``
        and ``time_bin_start`` are arrays, bins do not need to be contiguous.
        If this argument is provided, ``time_bin_size`` should not be provided.
    n_bins : int, optional
        The number of bins to use. Defaults to the number needed to fit all
        the original points. If both ``time_bin_start`` and ``time_bin_size``
        are provided and are scalar values, this determines the total bins
        within that interval. If ``time_bin_start`` is an iterable, this
        parameter will be ignored.
    aggregate_func : callable, optional
        The function to use for combining points in the same bin. Defaults
        to an internal implementation of nanmean.

    Returns
    -------
    binned_time_series : :class:`~astropy.timeseries.BinnedTimeSeries`
        The downsampled time series.
    """
    if not isinstance(time_series, TimeSeries):
        raise TypeError("time_series should be a TimeSeries")

    if time_bin_size is not None and not isinstance(
        time_bin_size, (u.Quantity, TimeDelta)
    ):
        raise TypeError("'time_bin_size' should be a Quantity or a TimeDelta")

    if time_bin_start is not None and not isinstance(time_bin_start, (Time, TimeDelta)):
        time_bin_start = Time(time_bin_start)

    if time_bin_end is not None and not isinstance(time_bin_end, (Time, TimeDelta)):
        time_bin_end = Time(time_bin_end)

    # Use the table sorted by time
    ts_sorted = time_series.iloc[:]

    # If start time is not provided, it is assumed to be the start of the timeseries
    if time_bin_start is None:
        time_bin_start = ts_sorted.time[0]

    # Total duration of the timeseries is needed for determining either
    # `time_bin_size` or `nbins` in the case of scalar `time_bin_start`
    if time_bin_start.isscalar:
        time_duration = (ts_sorted.time[-1] - time_bin_start).sec

    if time_bin_size is None and time_bin_end is None:
        if time_bin_start.isscalar:
            if n_bins is None:
                raise TypeError(
                    "With single 'time_bin_start' either 'n_bins', "
                    "'time_bin_size' or time_bin_end' must be provided"
                )
            else:
                # `nbins` defaults to the number needed to fit all points
                time_bin_size = time_duration / n_bins * u.s
        else:
            time_bin_end = np.maximum(ts_sorted.time[-1], time_bin_start[-1])

    if time_bin_start.isscalar:
        if time_bin_size is not None:
            if time_bin_size.isscalar:
                # Determine the number of bins
                if n_bins is None:
                    bin_size_sec = time_bin_size.to_value(u.s)
                    n_bins = int(np.ceil(time_duration / bin_size_sec))
        elif time_bin_end is not None:
            if not time_bin_end.isscalar:
                # Convert start time to an array and populate using `time_bin_end`
                scalar_start_time = time_bin_start
                time_bin_start = time_bin_end.replicate(copy=True)
                time_bin_start[0] = scalar_start_time
                time_bin_start[1:] = time_bin_end[:-1]

    # Check for overlapping bins, and warn if they are present
    if time_bin_end is not None:
        if (
            not time_bin_end.isscalar
            and not time_bin_start.isscalar
            and np.any(time_bin_start[1:] < time_bin_end[:-1])
        ):
            warnings.warn(
                "Overlapping bins should be avoided since they "
                "can lead to double-counting of data during binning.",
                AstropyUserWarning,
            )

    binned = BinnedTimeSeries(
        time_bin_size=time_bin_size,
        time_bin_start=time_bin_start,
        time_bin_end=time_bin_end,
        n_bins=n_bins,
    )

    if aggregate_func is None:
        aggregate_func = nanmean_reduceat

    # Start and end times of the binned timeseries
    bin_start = binned.time_bin_start
    bin_end = binned.time_bin_end

    # Set `n_bins` to match the length of `time_bin_start` if
    # `n_bins` is unspecified or if `time_bin_start` is an iterable
    if n_bins is None or not time_bin_start.isscalar:
        n_bins = len(bin_start)

    # Find the subset of the table that is inside the union of all bins
    # - output: `keep` a mask to create the subset
    # - use relative time in seconds `np.longdouble`` in in creating `keep` to speed up
    #   (`Time` object comparison is rather slow)
    # - tiny sacrifice on precision (< 0.01ns on 64 bit platform)
    rel_base = ts_sorted.time[0]
    rel_bin_start = _to_relative_longdouble(bin_start, rel_base)
    rel_bin_end = _to_relative_longdouble(bin_end, rel_base)
    rel_ts_sorted_time = _to_relative_longdouble(ts_sorted.time, rel_base)
    keep = (rel_ts_sorted_time >= rel_bin_start[0]) & (
        rel_ts_sorted_time <= rel_bin_end[-1]
    )

    # Find out indices to be removed because of noncontiguous bins
    #
    # Only need to check when adjacent bins have gaps, i.e.,
    # bin_start[ind + 1] > bin_end[ind]
    # - see: https://github.com/astropy/astropy/issues/13058#issuecomment-1090846697
    #   on thoughts on how to reduce the number of times to loop
    noncontiguous_bins_indices = np.where(rel_bin_start[1:] > rel_bin_end[:-1])[0]
    for ind in noncontiguous_bins_indices:
        delete_indices = np.where(
            np.logical_and(
                rel_ts_sorted_time > rel_bin_end[ind],
                rel_ts_sorted_time < rel_bin_start[ind + 1],
            )
        )
        keep[delete_indices] = False

    rel_subset_time = rel_ts_sorted_time[keep]

    # Figure out which bin each row falls in by sorting with respect
    # to the bin end times
    indices = np.searchsorted(rel_bin_end, rel_subset_time)

    # For time == bin_start[i+1] == bin_end[i], let bin_start takes precedence
    if len(indices) and np.all(rel_bin_start[1:] >= rel_bin_end[:-1]):
        indices_start = np.searchsorted(
            rel_subset_time, rel_bin_start[rel_bin_start <= rel_ts_sorted_time[-1]]
        )
        indices[indices_start] = np.arange(len(indices_start))

    # Determine rows where values are defined
    if len(indices):
        groups = np.hstack([0, np.nonzero(np.diff(indices))[0] + 1])
    else:
        groups = np.array([])

    # Find unique indices to determine which rows in the final time series
    # will not be empty.
    unique_indices = np.unique(indices)

    # Add back columns
    subset = ts_sorted[keep]
    for colname in subset.colnames:
        if colname == "time":
            continue

        values = subset[colname]

        # FIXME: figure out how to avoid the following, if possible
        if not isinstance(values, (np.ndarray, u.Quantity)):
            warnings.warn(
                "Skipping column {0} since it has a mix-in type", AstropyUserWarning
            )
            continue

        # TODO: This was written before MaskedQuantity were possible.
        # Should we return that by default, instead of using np.nan?
        if isinstance(values, u.Quantity):
            data = np.full_like(values, np.nan, shape=(n_bins,))
            # Pass ndarray (`values.value`) instead of Quantity (`values`)
            # reduceat() to avoid significant performance hit for cases
            # aggregate_func does not have optimized reduceat, e.g., np.nanmean.
            data[unique_indices] = u.Quantity(
                reduceat(values.value, groups, aggregate_func), values.unit, copy=False
            )
        else:
            data = np.ma.zeros(n_bins, dtype=values.dtype)
            data[unique_indices] = reduceat(values, groups, aggregate_func)

        if hasattr(data, "mask"):
            omitted = np.ones(data.shape, bool)
            omitted[unique_indices] = False
            data.mask |= omitted

        binned[colname] = data

    return binned



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module tests some methods related to ``CDS`` format
reader/writer.
Requires `pyyaml <https://pyyaml.org/>`_ to be installed.
"""

from io import StringIO

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import ascii
from astropy.table import Column, MaskedColumn, QTable, Table
from astropy.time import Time
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyWarning

test_dat = [
    "names e d s i",
    "HD81809 1E-7 22.25608 +2 67",
    "HD103095 -31.6e5 +27.2500 -9E34 -30",
]


def test_roundtrip_mrt_table():
    """
    Tests whether or not the CDS writer can roundtrip a table,
    i.e. read a table to ``Table`` object and write it exactly
    as it is back to a file. Since, presently CDS uses a
    MRT format template while writing, only the Byte-By-Byte
    and the data section of the table can be compared between
    original and the newly written table.

    Further, the CDS Reader does not have capability to recognize
    column format from the header of a CDS/MRT table, so this test
    can work for a limited set of simple tables, which don't have
    whitespaces in the column values or mix-in columns. Because of
    this the written table output cannot be directly matched with
    the original file and have to be checked against a list of lines.
    Masked columns are read properly though, and thus are being tested
    during round-tripping.

    The difference between ``cdsFunctional2.dat`` file and ``exp_output``
    is the following:
        * Metadata is different because MRT template is used for writing.
        * Spacing between ``Label`` and ``Explanations`` column in the
            Byte-By-Byte.
        * Units are written as ``[cm.s-2]`` and not ``[cm/s2]``, since both
            are valid according to CDS/MRT standard.
    """
    exp_output = [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 7  A7       ---    ID       Star ID                              ",
        " 9-12  I4       K      Teff     [4337/4654] Effective temperature    ",
        "14-17  F4.2   [cm.s-2] logg     [0.77/1.28] Surface gravity          ",
        "19-22  F4.2     km.s-1 vturb    [1.23/1.82] Micro-turbulence velocity",
        "24-28  F5.2     [-]    [Fe/H]   [-2.11/-1.5] Metallicity             ",
        "30-33  F4.2     [-]    e_[Fe/H] ? rms uncertainty on [Fe/H]          ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "S05-5   4337 0.77 1.80 -2.07     ",
        "S08-229 4625 1.23 1.23 -1.50     ",
        "S05-10  4342 0.91 1.82 -2.11 0.14",
        "S05-47  4654 1.28 1.74 -1.64 0.16",
    ]
    dat = get_pkg_data_filename(
        "data/cdsFunctional2.dat", package="astropy.io.ascii.tests"
    )
    t = Table.read(dat, format="ascii.mrt")
    out = StringIO()
    t.write(out, format="ascii.mrt")
    lines = out.getvalue().splitlines()
    i_bbb = lines.index("=" * 80)
    lines = lines[i_bbb:]  # Select Byte-By-Byte section and later lines.
    assert lines == exp_output


def test_write_byte_by_byte_units():
    t = ascii.read(test_dat)
    col_units = [None, u.C, u.kg, u.m / u.s, u.year]
    t._set_column_attribute("unit", col_units)
    # Add a column with magnitude units.
    # Note that magnitude has to be assigned for each value explicitly.
    t["magnitude"] = [u.Magnitude(25), u.Magnitude(-9)]
    col_units.append(u.mag)
    out = StringIO()
    t.write(out, format="ascii.mrt")
    # Read written table.
    columns = ascii.read(out.getvalue(), format="cds").itercols()
    assert [col.unit for col in columns] == col_units


def test_write_readme_with_default_options():
    exp_output = [
        "Title:",
        "Authors:",
        "Table:",
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 8  A8     ---    names   Description of names              ",
        "10-14  E5.1   ---    e       [-3160000.0/0.01] Description of e",
        "16-23  F8.5   ---    d       [22.25/27.25] Description of d    ",
        "25-31  E7.1   ---    s       [-9e+34/2.0] Description of s     ",
        "33-35  I3     ---    i       [-30/67] Description of i         ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD81809  1e-07  22.25608   2e+00  67",
        "HD103095 -3e+06 27.25000  -9e+34 -30",
    ]
    t = ascii.read(test_dat)
    out = StringIO()
    t.write(out, format="ascii.mrt")
    assert out.getvalue().splitlines() == exp_output


def test_write_empty_table():
    out = StringIO()
    with pytest.raises(NotImplementedError):
        Table().write(out, format="ascii.mrt")


def test_write_null_data_values():
    exp_output = [
        "HD81809  1e-07  22.25608  2.0e+00  67",
        "HD103095 -3e+06 27.25000 -9.0e+34 -30",
        "Sun                       5.3e+27    ",
    ]
    t = ascii.read(test_dat)
    t.add_row(
        ["Sun", "3.25", "0", "5.3e27", "2"], mask=[False, True, True, False, True]
    )
    out = StringIO()
    t.write(out, format="ascii.mrt")
    lines = out.getvalue().splitlines()
    i_secs = [i for i, s in enumerate(lines) if s.startswith(("------", "======="))]
    lines = lines[i_secs[-1] + 1 :]  # Last section is the data.
    assert lines == exp_output


def test_write_byte_by_byte_for_masked_column():
    """
    This test differs from the ``test_write_null_data_values``
    above in that it tests the column value limits in the Byte-By-Byte
    description section for columns whose values are masked.
    It also checks the description for columns with same values.
    """
    exp_output = [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 8  A8     ---    names   Description of names          ",
        "10-14  E5.1   ---    e       [0.0/0.01]? Description of e  ",
        "16-17  F2.0   ---    d       ? Description of d            ",
        "19-25  E7.1   ---    s       [-9e+34/2.0] Description of s ",
        "27-29  I3     ---    i       [-30/67] Description of i     ",
        "31-33  F3.1   ---    sameF   [5.0/5.0] Description of sameF",
        "35-36  I2     ---    sameI   [20] Description of sameI     ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD81809  1e-07    2e+00  67 5.0 20",
        "HD103095         -9e+34 -30 5.0 20",
    ]
    t = ascii.read(test_dat)
    t.add_column([5.0, 5.0], name="sameF")
    t.add_column([20, 20], name="sameI")
    t["e"] = MaskedColumn(t["e"], mask=[False, True])
    t["d"] = MaskedColumn(t["d"], mask=[True, True])
    out = StringIO()
    t.write(out, format="ascii.mrt")
    lines = out.getvalue().splitlines()
    i_bbb = lines.index("=" * 80)
    lines = lines[i_bbb:]  # Select Byte-By-Byte section and later lines.
    assert lines == exp_output


exp_coord_cols_output = {
    "generic": [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 8  A8     ---    names   Description of names              ",
        "10-14  E5.1   ---    e       [-3160000.0/0.01] Description of e",
        "16-23  F8.5   ---    d       [22.25/27.25] Description of d    ",
        "25-31  E7.1   ---    s       [-9e+34/2.0] Description of s     ",
        "33-35  I3     ---    i       [-30/67] Description of i         ",
        "37-39  F3.1   ---    sameF   [5.0/5.0] Description of sameF    ",
        "41-42  I2     ---    sameI   [20] Description of sameI         ",
        "44-45  I2     h      RAh     Right Ascension (hour)            ",
        "47-48  I2     min    RAm     Right Ascension (minute)          ",
        "50-62  F13.10 s      RAs     Right Ascension (second)          ",
        "   64  A1     ---    DE-     Sign of Declination               ",
        "65-66  I2     deg    DEd     Declination (degree)              ",
        "68-69  I2     arcmin DEm     Declination (arcmin)              ",
        "71-82  F12.9  arcsec DEs     Declination (arcsec)              ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD81809  1e-07  22.25608   2e+00  67 5.0 20 22 02 15.4500000000 -61 39 34.599996000",
        "HD103095 -3e+06 27.25000  -9e+34 -30 5.0 20 12 48 15.2244072000 +17 46 26.496624000",
    ],
    "positive_de": [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 8  A8     ---    names   Description of names              ",
        "10-14  E5.1   ---    e       [-3160000.0/0.01] Description of e",
        "16-23  F8.5   ---    d       [22.25/27.25] Description of d    ",
        "25-31  E7.1   ---    s       [-9e+34/2.0] Description of s     ",
        "33-35  I3     ---    i       [-30/67] Description of i         ",
        "37-39  F3.1   ---    sameF   [5.0/5.0] Description of sameF    ",
        "41-42  I2     ---    sameI   [20] Description of sameI         ",
        "44-45  I2     h      RAh     Right Ascension (hour)            ",
        "47-48  I2     min    RAm     Right Ascension (minute)          ",
        "50-62  F13.10 s      RAs     Right Ascension (second)          ",
        "   64  A1     ---    DE-     Sign of Declination               ",
        "65-66  I2     deg    DEd     Declination (degree)              ",
        "68-69  I2     arcmin DEm     Declination (arcmin)              ",
        "71-82  F12.9  arcsec DEs     Declination (arcsec)              ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD81809  1e-07  22.25608   2e+00  67 5.0 20 12 48 15.2244072000 +17 46 26.496624000",
        "HD103095 -3e+06 27.25000  -9e+34 -30 5.0 20 12 48 15.2244072000 +17 46 26.496624000",
    ],
    "galactic": [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 8  A8     ---    names   Description of names              ",
        "10-14  E5.1   ---    e       [-3160000.0/0.01] Description of e",
        "16-23  F8.5   ---    d       [22.25/27.25] Description of d    ",
        "25-31  E7.1   ---    s       [-9e+34/2.0] Description of s     ",
        "33-35  I3     ---    i       [-30/67] Description of i         ",
        "37-39  F3.1   ---    sameF   [5.0/5.0] Description of sameF    ",
        "41-42  I2     ---    sameI   [20] Description of sameI         ",
        "44-59  F16.12 deg    GLON    Galactic Longitude                ",
        "61-76  F16.12 deg    GLAT    Galactic Latitude                 ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD81809  1e-07  22.25608   2e+00  67 5.0 20 330.071639591690 -45.548080484609",
        "HD103095 -3e+06 27.25000  -9e+34 -30 5.0 20 330.071639591690 -45.548080484609",
    ],
    "ecliptic": [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 8  A8     ---    names   Description of names                       ",
        "10-14  E5.1   ---    e       [-3160000.0/0.01] Description of e         ",
        "16-23  F8.5   ---    d       [22.25/27.25] Description of d             ",
        "25-31  E7.1   ---    s       [-9e+34/2.0] Description of s              ",
        "33-35  I3     ---    i       [-30/67] Description of i                  ",
        "37-39  F3.1   ---    sameF   [5.0/5.0] Description of sameF             ",
        "41-42  I2     ---    sameI   [20] Description of sameI                  ",
        "44-59  F16.12 deg    ELON    Ecliptic Longitude (geocentrictrueecliptic)",
        "61-76  F16.12 deg    ELAT    Ecliptic Latitude (geocentrictrueecliptic) ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD81809  1e-07  22.25608   2e+00  67 5.0 20 306.224208650096 -45.621789850825",
        "HD103095 -3e+06 27.25000  -9e+34 -30 5.0 20 306.224208650096 -45.621789850825",
    ],
}


def test_write_coord_cols():
    """
    There can only be one such coordinate column in a single table,
    because division of columns into individual component columns requires
    iterating over the table columns, which will have to be done again
    if additional such coordinate columns are present.
    """
    t = ascii.read(test_dat)
    t.add_column([5.0, 5.0], name="sameF")
    t.add_column([20, 20], name="sameI")

    # Coordinates of ASASSN-15lh
    coord = SkyCoord(330.564375, -61.65961111, unit=u.deg)
    # Coordinates of ASASSN-14li
    coordp = SkyCoord(192.06343503, 17.77402684, unit=u.deg)
    cols = [
        Column([coord, coordp]),  # Generic coordinate column
        coordp,  # Coordinate column with positive DEC
        coord.galactic,  # Galactic coordinates
        coord.geocentrictrueecliptic,  # Ecliptic coordinates
    ]

    # Loop through different types of coordinate columns.
    for col, coord_type in zip(cols, exp_coord_cols_output):
        exp_output = exp_coord_cols_output[coord_type]
        t["coord"] = col
        out = StringIO()
        t.write(out, format="ascii.mrt")
        lines = out.getvalue().splitlines()
        i_bbb = lines.index("=" * 80)
        lines = lines[i_bbb:]  # Select Byte-By-Byte section and later lines.
        # Check the written table.
        assert lines == exp_output

        # Check if the original table columns remains unmodified.
        assert t.colnames == ["names", "e", "d", "s", "i", "sameF", "sameI", "coord"]


def test_write_byte_by_byte_bytes_col_format():
    """
    Tests the alignment of Byte counts with respect to hyphen
    in the Bytes column of Byte-By-Byte. The whitespace around the
    hyphen is govered by the number of digits in the total Byte
    count. Single Byte columns should have a single Byte count
    without the hyphen.
    """
    exp_output = [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 8  A8     ---    names         Description of names              ",
        "10-21  E12.6  ---    e             [-3160000.0/0.01] Description of e",
        "23-30  F8.5   ---    d             [22.25/27.25] Description of d    ",
        "32-38  E7.1   ---    s             [-9e+34/2.0] Description of s     ",
        "40-42  I3     ---    i             [-30/67] Description of i         ",
        "44-46  F3.1   ---    sameF         [5.0/5.0] Description of sameF    ",
        "48-49  I2     ---    sameI         [20] Description of sameI         ",
        "   51  I1     ---    singleByteCol [2] Description of singleByteCol  ",
        "53-54  I2     h      RAh           Right Ascension (hour)            ",
        "56-57  I2     min    RAm           Right Ascension (minute)          ",
        "59-71  F13.10 s      RAs           Right Ascension (second)          ",
        "   73  A1     ---    DE-           Sign of Declination               ",
        "74-75  I2     deg    DEd           Declination (degree)              ",
        "77-78  I2     arcmin DEm           Declination (arcmin)              ",
        "80-91  F12.9  arcsec DEs           Declination (arcsec)              ",
        "--------------------------------------------------------------------------------",
    ]
    t = ascii.read(test_dat)
    t.add_column([5.0, 5.0], name="sameF")
    t.add_column([20, 20], name="sameI")
    t["coord"] = SkyCoord(330.564375, -61.65961111, unit=u.deg)
    t["singleByteCol"] = [2, 2]
    t["e"].format = ".5E"
    out = StringIO()
    t.write(out, format="ascii.mrt")
    lines = out.getvalue().splitlines()
    i_secs = [i for i, s in enumerate(lines) if s.startswith(("------", "======="))]
    # Select only the Byte-By-Byte section.
    lines = lines[i_secs[0] : i_secs[-2]]
    lines.append("-" * 80)  # Append a separator line.
    assert lines == exp_output


def test_write_byte_by_byte_wrapping():
    """
    Test line wrapping in the description column of the
    Byte-By-Byte section of the ReadMe.
    """
    exp_output = """\
================================================================================
Byte-by-byte Description of file: table.dat
--------------------------------------------------------------------------------
 Bytes Format Units  Label     Explanations
--------------------------------------------------------------------------------
 1- 8  A8     ---    thisIsALongColumnLabel This is a tediously long
                                           description. But they do sometimes
                                           have them. Better to put extra
                                           details in the notes. This is a
                                           tediously long description. But they
                                           do sometimes have them. Better to put
                                           extra details in the notes.
10-14  E5.1   ---    e                      [-3160000.0/0.01] Description of e
16-23  F8.5   ---    d                      [22.25/27.25] Description of d
--------------------------------------------------------------------------------
"""
    t = ascii.read(test_dat)
    t.remove_columns(["s", "i"])
    description = (
        "This is a tediously long description."
        " But they do sometimes have them."
        " Better to put extra details in the notes. "
    )
    t["names"].description = description * 2
    t["names"].name = "thisIsALongColumnLabel"
    out = StringIO()
    t.write(out, format="ascii.mrt")
    lines = out.getvalue().splitlines()
    i_secs = [i for i, s in enumerate(lines) if s.startswith(("------", "======="))]
    # Select only the Byte-By-Byte section.
    lines = lines[i_secs[0] : i_secs[-2]]
    lines.append("-" * 80)  # Append a separator line.
    assert lines == exp_output.splitlines()


def test_write_mixin_and_broken_cols():
    """
    Tests conversion to string values for ``mix-in`` columns other than
    ``SkyCoord`` and for columns with only partial ``SkyCoord`` values.
    """
    exp_output = [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        "  1-  7  A7     ---    name    Description of name       ",
        "  9- 74  A66    ---    Unknown Description of Unknown    ",
        " 76-114  A39    ---    cart    Description of cart       ",
        "116-138  A23    ---    time    Description of time       ",
        "140-142  F3.1   m      q       [1.0/1.0] Description of q",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD81809 <SkyCoord (ICRS): (ra, dec) in deg",
        "    (330.564375, -61.65961111)> (0.41342785, -0.23329341, -0.88014294)  2019-01-01 00:00:00.000 1.0",
        "random  12                                                                 (0.41342785, -0.23329341, -0.88014294)  2019-01-01 00:00:00.000 1.0",
    ]
    t = Table()
    t["name"] = ["HD81809"]
    coord = SkyCoord(330.564375, -61.65961111, unit=u.deg)
    t["coord"] = Column(coord)
    t.add_row(["random", 12])
    t["cart"] = coord.cartesian
    t["time"] = Time("2019-1-1")
    t["q"] = u.Quantity(1.0, u.m)
    out = StringIO()
    t.write(out, format="ascii.mrt")
    lines = out.getvalue().splitlines()
    i_bbb = lines.index("=" * 80)
    lines = lines[i_bbb:]  # Select Byte-By-Byte section and later lines.
    # Check the written table.
    assert lines == exp_output


def test_write_extra_skycoord_cols():
    """
    Tests output for cases when table contains multiple ``SkyCoord`` columns.
    """
    exp_output = [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 7  A7     ---    name    Description of name     ",
        " 9-10  I2     h      RAh     Right Ascension (hour)  ",
        "12-13  I2     min    RAm     Right Ascension (minute)",
        "15-27  F13.10 s      RAs     Right Ascension (second)",
        "   29  A1     ---    DE-     Sign of Declination     ",
        "30-31  I2     deg    DEd     Declination (degree)    ",
        "33-34  I2     arcmin DEm     Declination (arcmin)    ",
        "36-47  F12.9  arcsec DEs     Declination (arcsec)    ",
        "49-62  A14    ---    coord2  Description of coord2   ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD4760   0 49 39.9000000000 +06 24 07.999200000 12.4163 6.407 ",
        "HD81809 22 02 15.4500000000 -61 39 34.599996000 330.564 -61.66",
    ]
    t = Table()
    t["name"] = ["HD4760", "HD81809"]
    t["coord1"] = SkyCoord([12.41625, 330.564375], [6.402222, -61.65961111], unit=u.deg)
    t["coord2"] = SkyCoord([12.41630, 330.564400], [6.407, -61.66], unit=u.deg)
    out = StringIO()
    with pytest.warns(
        UserWarning,
        match=r"column 2 is being skipped with designation of a "
        r"string valued column `coord2`",
    ):
        t.write(out, format="ascii.mrt")

    lines = out.getvalue().splitlines()
    i_bbb = lines.index("=" * 80)
    lines = lines[i_bbb:]  # Select Byte-By-Byte section and following lines.
    # Check the written table.
    assert lines[:-2] == exp_output[:-2]

    for a, b in zip(lines[-2:], exp_output[-2:]):
        assert a[:18] == b[:18]
        assert a[30:42] == b[30:42]
        assert_allclose(np.fromstring(a[2:], sep=" "), np.fromstring(b[2:], sep=" "))


def test_write_skycoord_with_format():
    """
    Tests output with custom setting for ``SkyCoord`` (second) columns.
    """
    exp_output = [
        "================================================================================",
        "Byte-by-byte Description of file: table.dat",
        "--------------------------------------------------------------------------------",
        " Bytes Format Units  Label     Explanations",
        "--------------------------------------------------------------------------------",
        " 1- 7  A7     ---    name    Description of name     ",
        " 9-10  I2     h      RAh     Right Ascension (hour)  ",
        "12-13  I2     min    RAm     Right Ascension (minute)",
        "15-19  F5.2   s      RAs     Right Ascension (second)",
        "   21  A1     ---    DE-     Sign of Declination     ",
        "22-23  I2     deg    DEd     Declination (degree)    ",
        "25-26  I2     arcmin DEm     Declination (arcmin)    ",
        "28-31  F4.1   arcsec DEs     Declination (arcsec)    ",
        "--------------------------------------------------------------------------------",
        "Notes:",
        "--------------------------------------------------------------------------------",
        "HD4760   0 49 39.90 +06 24 08.0",
        "HD81809 22 02 15.45 -61 39 34.6",
    ]
    t = Table()
    t["name"] = ["HD4760", "HD81809"]
    t["coord"] = SkyCoord([12.41625, 330.564375], [6.402222, -61.65961111], unit=u.deg)

    out = StringIO()
    # This will raise a warning because `formats` is checked before the writer creating the
    # final list of columns is called.
    with pytest.warns(
        AstropyWarning,
        match=r"The key.s. {'[RD][AE]s', '[RD][AE]s'} specified in "
        r"the formats argument do not match a column name.",
    ):
        t.write(out, format="ascii.mrt", formats={"RAs": "05.2f", "DEs": "04.1f"})

    lines = out.getvalue().splitlines()
    i_bbb = lines.index("=" * 80)
    lines = lines[i_bbb:]  # Select Byte-By-Byte section and following lines.
    # Check the written table.
    assert lines == exp_output


def test_write_qtable():
    # Regression test for gh-12804
    qt = QTable([np.arange(4) * u.m, ["a", "b", "c", "ddd"]], names=["a", "b"])
    out = StringIO()
    qt.write(out, format="mrt")
    result = out.getvalue()
    assert "Description of a" in result
    assert "Description of b" in result



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This file was automatically generated from ply. To re-generate this file,
# remove it from this folder, then build astropy and run the tests in-place:
#
#   python setup.py build_ext --inplace
#   pytest astropy/units
#
# You can then commit the changes to this file.

# ogip_lextab.py. This file automatically created by PLY (version 3.11). Don't edit!
_tabversion   = '3.10'
_lextokens    = set(('CLOSE_PAREN', 'DIVISION', 'FLOAT', 'FUNCNAME', 'INT', 'LIT10', 'OPEN_PAREN', 'POWER', 'STAR', 'UNIT', 'UNKNOWN', 'WHITESPACE'))
_lexreflags   = 64
_lexliterals  = ''
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_FLOAT>[+-]?((((\\d+\\.?\\d*)|(\\.\\d+))([eE][+-]?\\d+))|(((\\d+\\.\\d*)|(\\.\\d+))([eE][+-]?\\d+)?)))|(?P<t_INT>[+-]?\\d+)|(?P<t_LIT10>10)|(?P<t_UNKNOWN>[Uu][Nn][Kk][Nn][Oo][Ww][Nn])|(?P<t_FUNCNAME>((sqrt)|(ln)|(exp)|(log)|(sin)|(cos)|(tan)|(asin)|(acos)|(atan)|(sinh)|(cosh)|(tanh))(?=\\ *\\())|(?P<t_UNIT>[a-zA-Z][a-zA-Z_]*)|(?P<t_DIVISION>[ \t]*/[ \t]*)|(?P<t_WHITESPACE>[ \t]+)|(?P<t_POWER>\\*\\*)|(?P<t_OPEN_PAREN>\\()|(?P<t_CLOSE_PAREN>\\))|(?P<t_STAR>\\*)', [None, ('t_FLOAT', 'FLOAT'), None, None, None, None, None, None, None, None, None, None, None, ('t_INT', 'INT'), ('t_LIT10', 'LIT10'), ('t_UNKNOWN', 'UNKNOWN'), ('t_FUNCNAME', 'FUNCNAME'), None, None, None, None, None, None, None, None, None, None, None, None, None, None, ('t_UNIT', 'UNIT'), (None, 'DIVISION'), (None, 'WHITESPACE'), (None, 'POWER'), (None, 'OPEN_PAREN'), (None, 'CLOSE_PAREN'), (None, 'STAR')])]}
_lexstateignore = {'INITIAL': ''}
_lexstateerrorf = {'INITIAL': 't_error'}
_lexstateeoff = {}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This is a collection of monkey patches and workarounds for bugs in
earlier versions of Numpy.
"""

import numpy as np

from astropy.utils import minversion

__all__ = [
    "COPY_IF_NEEDED",
    "NUMPY_LT_1_25",
    "NUMPY_LT_1_26",
    "NUMPY_LT_2_0",
    "NUMPY_LT_2_1",
    "NUMPY_LT_2_2",
    "NUMPY_LT_2_3",
    "NUMPY_LT_2_4",
]

# TODO: It might also be nice to have aliases to these named for specific
# features/bugs we're checking for (ex:
# astropy.table.table._BROKEN_UNICODE_TABLE_SORT)
NUMPY_LT_1_25 = not minversion(np, "1.25")
NUMPY_LT_1_26 = not minversion(np, "1.26")
NUMPY_LT_2_0 = not minversion(np, "2.0")
NUMPY_LT_2_1 = not minversion(np, "2.1.0.dev")
NUMPY_LT_2_2 = not minversion(np, "2.2.0.dev0")
NUMPY_LT_2_3 = not minversion(np, "2.3.0.dev0")
NUMPY_LT_2_4 = not minversion(np, "2.4.0.dev0")


COPY_IF_NEEDED = False if NUMPY_LT_2_0 else None



# Download an example FITS file, create a 2D cutout, and save it to a
# new FITS file, including the updated cutout WCS.
from astropy.io import fits
from astropy.nddata import Cutout2D
from astropy.utils.data import download_file
from astropy.wcs import WCS


def download_image_save_cutout(url, position, size):
    # Download the image
    filename = download_file(url)

    # Load the image and the WCS
    hdu = fits.open(filename)[0]
    wcs = WCS(hdu.header)

    # Make the cutout, including the WCS
    cutout = Cutout2D(hdu.data, position=position, size=size, wcs=wcs)

    # Put the cutout image in the FITS HDU
    hdu.data = cutout.data

    # Update the FITS header with the cutout WCS
    hdu.header.update(cutout.wcs.to_header())

    # Write the cutout to a new FITS file
    cutout_filename = "example_cutout.fits"
    hdu.writeto(cutout_filename, overwrite=True)


if __name__ == "__main__":
    url = "https://astropy.stsci.edu/data/photometry/spitzer_example_image.fits"

    position = (500, 300)
    size = (400, 400)
    download_image_save_cutout(url, position, size)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.coordinates.attributes import EarthLocationAttribute, TimeAttribute
from astropy.coordinates.baseframe import BaseCoordinateFrame, base_doc
from astropy.coordinates.earth import EarthLocation
from astropy.coordinates.representation import (
    CartesianDifferential,
    CartesianRepresentation,
)
from astropy.utils.decorators import format_doc

from .utils import DEFAULT_OBSTIME, EARTH_CENTER

__all__ = ["ITRS"]

doc_footer = """
    Other parameters
    ----------------
    obstime : `~astropy.time.Time`
        The time at which the observation is taken.  Used for determining the
        position of the Earth and its precession.
    location : `~astropy.coordinates.EarthLocation`
        The location on the Earth.  This can be specified either as an
        `~astropy.coordinates.EarthLocation` object or as anything that can be
        transformed to an `~astropy.coordinates.ITRS` frame. The default is the
        centre of the Earth.
"""


@format_doc(base_doc, components="", footer=doc_footer)
class ITRS(BaseCoordinateFrame):
    """
    A coordinate or frame in the International Terrestrial Reference System
    (ITRS).  This is approximately a geocentric system, although strictly it is
    defined by a series of reference locations near the surface of the Earth (the ITRF).
    For more background on the ITRS, see the references provided in the
    :ref:`astropy:astropy-coordinates-seealso` section of the documentation.

    This frame also includes frames that are defined *relative* to the center of the Earth,
    but that are offset (in both position and velocity) from the center of the Earth. You
    may see such non-geocentric coordinates referred to as "topocentric".

    Topocentric ITRS frames are convenient for observations of near Earth objects where
    stellar aberration is not included. One can merely subtract the observing site's
    EarthLocation geocentric ITRS coordinates from the object's geocentric ITRS coordinates,
    put the resulting vector into a topocentric ITRS frame and then transform to
    `~astropy.coordinates.AltAz` or `~astropy.coordinates.HADec`. The other way around is
    to transform an observed `~astropy.coordinates.AltAz` or `~astropy.coordinates.HADec`
    position to a topocentric ITRS frame and add the observing site's EarthLocation geocentric
    ITRS coordinates to yield the object's geocentric ITRS coordinates.

    On the other hand, using ``transform_to`` to transform geocentric ITRS coordinates to
    topocentric ITRS, observed `~astropy.coordinates.AltAz`, or observed
    `~astropy.coordinates.HADec` coordinates includes the difference between stellar aberration
    from the point of view of an observer at the geocenter and stellar aberration from the
    point of view of an observer on the surface of the Earth. If the geocentric ITRS
    coordinates of the object include stellar aberration at the geocenter (e.g. certain ILRS
    ephemerides), then this is the way to go.

    Note to ILRS ephemeris users: Astropy does not currently consider relativistic
    effects of the Earth's gravatational field. Nor do the `~astropy.coordinates.AltAz`
    or `~astropy.coordinates.HADec` refraction corrections compute the change in the
    range due to the curved path of light through the atmosphere, so Astropy is no
    substitute for the ILRS software in these respects.

    """

    default_representation = CartesianRepresentation
    default_differential = CartesianDifferential

    obstime = TimeAttribute(
        default=DEFAULT_OBSTIME, doc="The reference time (e.g., time of observation)"
    )
    location = EarthLocationAttribute(
        default=EARTH_CENTER, doc="The location on Earth of the observer"
    )

    @property
    def earth_location(self):
        """
        The data in this frame as an `~astropy.coordinates.EarthLocation` class.
        """
        cart = self.represent_as(CartesianRepresentation)
        return EarthLocation(
            x=cart.x + self.location.x,
            y=cart.y + self.location.y,
            z=cart.z + self.location.z,
        )


# Self-transform is in intermediate_rotation_transforms.py with all the other
# ITRS transforms



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy import units as u
from astropy.coordinates import representation as r
from astropy.coordinates.baseframe import (
    BaseCoordinateFrame,
    RepresentationMapping,
    base_doc,
)
from astropy.utils.decorators import format_doc

from .galactic import Galactic

__all__ = ["Supergalactic"]


doc_components = """
    sgl : `~astropy.coordinates.Angle`, optional, keyword-only
        The supergalactic longitude for this object (``sgb`` must also be given and
        ``representation`` must be None).
    sgb : `~astropy.coordinates.Angle`, optional, keyword-only
        The supergalactic latitude for this object (``sgl`` must also be given and
        ``representation`` must be None).
    distance : `~astropy.units.Quantity` ['speed'], optional, keyword-only
        The Distance for this object along the line-of-sight.

    pm_sgl_cossgb : `~astropy.units.Quantity` ['angular speed'], optional, keyword-only
        The proper motion in Right Ascension for this object (``pm_sgb`` must
        also be given).
    pm_sgb : `~astropy.units.Quantity` ['angular speed'], optional, keyword-only
        The proper motion in Declination for this object (``pm_sgl_cossgb`` must
        also be given).
    radial_velocity : `~astropy.units.Quantity` ['speed'], optional, keyword-only
        The radial velocity of this object.
"""


@format_doc(base_doc, components=doc_components, footer="")
class Supergalactic(BaseCoordinateFrame):
    """
    Supergalactic Coordinates
    (see Lahav et al. 2000, <https://ui.adsabs.harvard.edu/abs/2000MNRAS.312..166L>,
    and references therein).
    """

    frame_specific_representation_info = {
        r.SphericalRepresentation: [
            RepresentationMapping("lon", "sgl"),
            RepresentationMapping("lat", "sgb"),
        ],
        r.CartesianRepresentation: [
            RepresentationMapping("x", "sgx"),
            RepresentationMapping("y", "sgy"),
            RepresentationMapping("z", "sgz"),
        ],
        r.CartesianDifferential: [
            RepresentationMapping("d_x", "v_x", u.km / u.s),
            RepresentationMapping("d_y", "v_y", u.km / u.s),
            RepresentationMapping("d_z", "v_z", u.km / u.s),
        ],
    }

    default_representation = r.SphericalRepresentation
    default_differential = r.SphericalCosLatDifferential

    # North supergalactic pole in Galactic coordinates.
    # Needed for transformations to/from Galactic coordinates.
    _nsgp_gal = Galactic(l=47.37 * u.degree, b=+6.32 * u.degree)



from pathlib import Path
from threading import Lock

import pytest

from astropy import units as u
from astropy.config import set_temp_cache
from astropy.coordinates import EarthLocation, Latitude, Longitude, UnknownSiteException
from astropy.coordinates.sites import SiteRegistry, get_downloaded_sites
from astropy.tests.helper import assert_quantity_allclose

SITE_DATA_LOCK = Lock()


@pytest.fixture(scope="function")
def local_site_data(monkeypatch):
    with SITE_DATA_LOCK:
        monkeypatch.setattr(EarthLocation, "_site_registry", None)
        with set_temp_cache(Path(__file__).with_name("data") / "sites"):
            yield


@pytest.mark.remote_data(source="astropy")
def test_online_sites():
    reg = get_downloaded_sites()

    keck = reg["keck"]
    lon, lat, el = keck.to_geodetic()
    assert_quantity_allclose(
        lon, -Longitude("155:28.7", unit=u.deg), atol=0.001 * u.deg
    )
    assert_quantity_allclose(lat, Latitude("19:49.7", unit=u.deg), atol=0.001 * u.deg)
    assert_quantity_allclose(el, 4160 * u.m, atol=1 * u.m)

    names = reg.names
    assert "keck" in names
    assert "ctio" in names

    # The JSON file contains `name` and `aliases` for each site, and astropy
    # should use names from both, but not empty strings [#12721].
    assert "" not in names
    assert "Royal Observatory Greenwich" in names

    with pytest.raises(
        KeyError,
        match="Site 'nonexistent' not in database. Use the 'names' attribute to see",
    ):
        reg["nonexistent"]

    with pytest.raises(
        KeyError,
        match="Site 'kec' not in database. Use the 'names' attribute to see available",
    ):
        reg["kec"]


@pytest.mark.usefixtures("local_site_data")
def test_EarthLocation_basic():
    greenwichel = EarthLocation.of_site("greenwich")
    lon, lat, el = greenwichel.to_geodetic()
    assert_quantity_allclose(lon, Longitude("0:0:0", unit=u.deg), atol=10 * u.arcsec)
    assert_quantity_allclose(lat, Latitude("51:28:40", unit=u.deg), atol=1 * u.arcsec)
    assert_quantity_allclose(el, 46 * u.m, atol=1 * u.m)

    names = EarthLocation.get_site_names()
    assert "greenwich" in names
    assert "example_site" in names

    with pytest.raises(
        KeyError,
        match="Site 'nonexistent' not in database. Use EarthLocation.get_site_names",
    ):
        EarthLocation.of_site("nonexistent")

    with pytest.raises(TypeError, match="^site name None is not a 'str'$"):
        EarthLocation.of_site(None)


@pytest.mark.parametrize(
    "class_method,args",
    [(EarthLocation.get_site_names, []), (EarthLocation.of_site, ["greenwich"])],
)
def test_Earthlocation_refresh_cache_is_mandatory_kwarg(class_method, args):
    with pytest.raises(
        TypeError,
        match=(
            rf".*{class_method.__name__}\(\) takes [12] positional "
            "arguments? but [23] were given$"
        ),
    ):
        class_method(*args, False)


@pytest.mark.xfail(reason="regression test for #18572", raises=AssertionError)
@pytest.mark.usefixtures("local_site_data")
def test_Earthlocation_get_site_names_refresh_cache():
    assert len(EarthLocation.get_site_names()) == 3
    # Regression test for #18572 - refresh_cache=True had no effect
    assert len(EarthLocation.get_site_names(refresh_cache=True)) > 3


@pytest.mark.usefixtures("local_site_data")
def test_Earthlocation_of_site_no_refresh_cache():
    with pytest.raises(UnknownSiteException, match="^Site 'keck' not in database"):
        EarthLocation.of_site("keck")


@pytest.mark.xfail(reason="regression test for #18572", raises=UnknownSiteException)
@pytest.mark.usefixtures("local_site_data")
def test_Earthlocation_of_site_refresh_cache():
    # Regression test for #18572 - refresh_cache=True had no effect
    keck = EarthLocation.of_site("keck", refresh_cache=True)
    assert_quantity_allclose(keck.lon, -155.47833333 * u.deg)


def test_registry():
    reg = SiteRegistry()
    assert len(reg.names) == 0

    loc = EarthLocation.from_geodetic(lat=1 * u.deg, lon=2 * u.deg, height=3 * u.km)
    reg.add_site(["sitea", "site A"], loc)
    assert len(reg.names) == 2
    assert reg["SIteA"] is loc
    assert reg["sIte a"] is loc


@pytest.mark.usefixtures("local_site_data")
def test_non_EarthLocation():
    """
    A regression test for a typo bug pointed out at the bottom of
    https://github.com/astropy/astropy/pull/4042
    """

    class EarthLocation2(EarthLocation):
        pass

    el2 = EarthLocation2.of_site("greenwich")
    assert type(el2) is EarthLocation2
    assert el2.info.name == "Royal Observatory Greenwich"


@pytest.mark.usefixtures("local_site_data")
def test_meta_present():
    assert (
        EarthLocation.of_site("greenwich").info.meta["source"]
        == "Ordnance Survey via http://gpsinformation.net/main/greenwich.htm and UNESCO"
    )
